"errors"

from enum import IntEnum
from typing import Optional, Union

from ark.core._api.deprecation import deprecated
from ark.core.idl.common_protocol import Error
from ark.core.utils import context


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.core.utils.errorsv3.ErrorCode",
)
class GptCode(IntEnum):
    http_code: int
    description: Optional[str]

    def __new__(
        cls, code_n: int, http_code: int = 500, description: str = ""
    ) -> "GptCode":
        obj = int.__new__(cls, code_n)
        obj._value_ = code_n

        obj.http_code = http_code
        obj.description = description
        return obj

    SignatureDoesNotMatch = (
        1709801,
        401,
        "Request signature we calculated does not match the signature you provided",
    )
    RequestTimeout = (
        1709802,
        408,
        "Request to service is timeout: it may happen on service is meeting heavy load",
    )
    InferenceServiceConnetionTimeout = (
        1709803,
        504,
        "Connection to service is timeout",
    )
    MissingAuthenticationHeader = (
        1709804,
        401,
        "Request is missing authentication header",
    )
    AuthenticationHeaderIsInvalid = (
        1709805,
        401,
        "Request authentication header content is invalid",
    )
    InternalServiceError = (1709806, 500, "Service has some internal Error")
    MissingParameter = 1709807, 400, "The request is missing parameter"
    InvalidParameter = (
        1709808,
        400,
        "A parameter specified in the request is not valid",
    )
    AuthenticationExpire = (1709809, 401, "You are failed to authentication")
    EndpointIsInvalid = (1709810, 400, "The specific endpoint is not exist or invalid")
    EndpointIsNotEnable = (1709811, 503, "The specific endpoint is closed")
    ChatNotSupportStreamMode = (1709812, 400, "Model not support stream mode")
    ReqTextExistRisk = (
        1709813,
        400,
        "The input text may contain sensitive information",
    )
    RespTextExistRisk = (
        1709814,
        400,
        "The output text may contain sensitive information",
    )
    EndpointRateLimitExceeded = 1709815, 429, "The request is throttled"
    ServiceConnectionRefused = (1709816, 503, "Connection to service is refused")
    ServiceConnectionClosed = (1709817, 502, "Connection to service is closed")
    UnauthorizedUserForEndpoint = (1709818, 401, "Failed to access endpoint")
    InvalidEndpointWithNoURL = (1709819, 401, "Endpoint is invalid")
    EndpointAccountRpmRateLimitExceeded = (
        1709820,
        429,
        "RPM (Requests Per Minute) limit of the account is exceeded",
    )
    EndpointAccountTpmRateLimitExceeded = (
        1709821,
        429,
        "TPM (Tokens Per Minute) limit of the account is exceeded",
    )
    ServiceResourceWaitQueueFull = (
        1709822,
        412,
        "Service has no more resource for more request put into queue",
    )
    EndpointIsPending = (1709823, 503, "The specific endpoint is pending")
    ServiceNotOpen = (1709824, 503, "The free trial quota has been consumed")
    MaasPlatformNotOpen = (
        1709825,
        503,
        "The user has not opened the maas platform service",
    )
    LowPriorityLimitExceeded = (
        1709826,
        429,
        "The request is throttled by low priority limiter",
    )
    TextCheckServiceError = (1709827, 500, "Fail to request business security service")
    APINotSupport = (1709828, 405, "API not supported")
    ModelNotSupportEmbeddings = (1709829, 405, "Model not support embeddings")
    UnauthorizedClientCertificate = (1709901, 401, "You are failed to authentication")
    InternalServiceShowMsgError = (1709902, 500, "Service has some internal")
    ChatNotSupportBatchStreamMode = (
        1709903,
        405,
        "Model not support batch stream mode",
    )
    ChatNotSupportBatchMode = (1709904, 405, "Model not support batch mode")
    RequestCanceled = (1709905, 500, "The request is canceled before response")
    NeedRetryError = (1709906, 500, "The request can retry afterwards")
    NoAccountIdError = (1709907, 500, "No AccountId in context")
    NoUserIdError = (1709908, 500, "No UserId in context")

    Unknown = (1709600, 500, "Unknown error")
    EngineInternalServiceError = (1709601, 500, "Engine internal service error")
    FunctionCallPostProcessError = (1709602, 500, "Function call post process error")
    ModelLoadingError = (
        1709912,
        503,
        "loading model, please try again later, or contact with administrator",
    )

    ImageSizeLimitExceeded = (
        1709830,
        400,
        (
            "The Image size exceeds limit, please reduce the image size, "
            "or contact with administrator"
        ),
    )
    ImageFormatNotSupported = (
        1709831,
        400,
        (
            "The Image format not supported, only (jpg, png), please modify image, "
            "or contact with administrator"
        ),
    )
    OnlyOneImageSupported = (
        1709832,
        400,
        (
            "Only one image supported in a session, please modify image, "
            "or contact with administrator"
        ),
    )
    OnlyOneMultiContentSupported = (
        1709833,
        400,
        (
            "Only one content supported in a session, please modify, "
            "or contact with administrator"
        ),
    )
    UrlUnSupported = (
        1709834,
        400,
        (
            "Only tos, http, https url supported, please modify, "
            "or contact with administrator"
        ),
    )
    OperationDenied = (
        1709835,
        403,
        ("Operation is denied."),
    )
    KnowledgeBaseError = (
        1709836,
        500,
        "Calling KnowledgeBase meet error",
    )
    ResourceNotFound = (
        1709837,
        404,
        "The specified resource is not found.",
    )
    AssistantRateLimitExceeded = (
        1709838,
        429,
        "Request limit of assistant is exceeded",
    )
    ActionRateLimitExceeded = (
        1709839,
        429,
        "Request limit of action is exceeded",
    )
    UnauthorizedUserForResource = (
        1709840,
        401,
        "Failed to access resource: not pass the authorization check.",
    )


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.core.utils.errorsv3.APIException",
)
class GPTException(Exception):
    def __init__(self, message: str, code_n: Union[GptCode, int]):
        super().__init__(message)
        self.message = message
        self.resource_id = context.get_resource_id()
        self.account_id = context.get_account_id() or ""

        if isinstance(code_n, GptCode):
            code_n = code_n
        else:
            try:
                code_n = GptCode(code_n)
            except ValueError:
                code_n = GptCode.Unknown

        self.code_n = code_n

    @property
    def code(self) -> str:
        return self.code_n.name

    @property
    def http_code(self) -> int:
        return self.code_n.http_code

    def __str__(self) -> str:
        return (
            "Detailed exception information is listed below.\n"
            + "account_id: {}\n"
            + "resource_id: {}\n"
            + "code_n: {}\n"
            + "code: {}\n"
            + "message: {}"
        ).format(
            self.account_id, self.resource_id, self.code_n, self.code, self.message
        )

    def to_error(self) -> Error:
        return Error(
            code=self.code,
            code_n=self.code_n.value,
            message=self.message,
        )


class InternalServiceError(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.EngineInternalServiceError)


class InvalidParameter(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.InvalidParameter)


class MissingParameter(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.MissingParameter)


class OperationDenied(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.OperationDenied)


class KnowledgeBaseError(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.KnowledgeBaseError)


class RequestTimeout(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.RequestTimeout)


class FunctionCallPostProcessError(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.FunctionCallPostProcessError)


class ReqTextExistRisk(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.ReqTextExistRisk)


class RespTextExistRisk(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.RespTextExistRisk)


class TextCheckServiceError(GPTException):
    def __init__(self, message: str):
        super().__init__(message, code_n=GptCode.TextCheckServiceError)


class RagException(Exception):
    def __init__(self, message: str, code_n: int):
        super().__init__(message)
        self.message = message
        self.code_n = code_n


class RagCode(IntEnum):
    description: Optional[str]

    def __new__(cls, code_n: int, description: Optional[str]) -> "RagCode":
        obj = int.__new__(cls, code_n)
        obj._value_ = code_n
        obj.description = description
        return obj

    FeishuParseFailed = (
        1709900,
        "Feishu parse failed",
    )


class FeishuParseException(RagException):
    def __init__(self, message: str):
        super().__init__(message, code_n=RagCode.FeishuParseFailed)



from ark.component.reader import (
    _reader,
    base_reader,
    common_reader,
    feishu_reader,
    struct_reader,
)

__all__ = [
    "_reader",
    "base_reader",
    "common_reader",
    "feishu_reader",
    "struct_reader",
]



import asyncio
from typing import Any, AsyncIterable, Optional


class AsyncTimedIterable:
    def __init__(
        self,
        iterable: AsyncIterable[Any],
        timeout: Optional[int] = None,
        sentinel: Optional[Any] = None,
    ):
        class AsyncTimedIterator:
            def __init__(self) -> None:
                self._iterator = iterable.__aiter__()

            async def __anext__(self) -> Any:
                try:
                    return await asyncio.wait_for(self._iterator.__anext__(), timeout)
                except asyncio.TimeoutError as e:
                    if sentinel:
                        raise sentinel
                    else:
                        raise e

        self._factory = AsyncTimedIterator

    def __aiter__(self) -> Any:
        return self._factory()



import asyncio
from typing import (
    Any,
    AsyncIterable,
    AsyncIterator,
    Coroutine,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
)

from ark.core.idl.common_protocol import ChatResponse


def merge_response(response: Iterator[ChatResponse]) -> Optional[ChatResponse]:
    if response is None:
        return None
    merged = ChatResponse.merge([i for i in response])
    return merged


def async_iterator_to_sync(loop: asyncio.AbstractEventLoop, gen: Any) -> Iterator[Any]:
    try:
        while True:
            i = asyncio.run_coroutine_threadsafe(gen.__anext__(), loop).result()
            yield i
    except StopAsyncIteration:
        pass


T = TypeVar("T")


def sync_iterator_to_async(
    loop: asyncio.AbstractEventLoop, gen: Iterator[T], _END: Any = None
) -> AsyncIterable[Any]:
    if _END is None:
        _END = object()

        def wrap(gen: Iterator[T]) -> Iterator[T]:
            yield from gen
            yield _END

        gen = wrap(gen)

    async def async_gen(
        loop: asyncio.AbstractEventLoop, gen: Iterator[T]
    ) -> AsyncIterable[T]:
        while True:
            i: T = await loop.run_in_executor(None, next, gen)
            if i is _END:
                break
            yield i

    return async_gen(loop, gen)


async def async_any(coroutines: Any) -> bool:
    done = asyncio.as_completed(coroutines)
    for task in done:
        if await task:
            return True
    return False


async def aenumerate(
    asequence: AsyncIterable[T], start: int = 0
) -> AsyncIterator[Tuple[int, T]]:
    """Asynchronously enumerate an async iterator from a given start value"""
    n = start
    async for elem in asequence:
        yield n, elem
        n += 1


async def anext(asequence: AsyncIterator[T]) -> T:
    """
    Asynchronously get next item from async generator
    Note: python3.10 has a anext builtin function
    """
    return await asequence.__anext__()


async def gather(*coros_or_futures: Coroutine) -> List[Any]:
    tasks: List[asyncio.Task[Any]] = [
        asyncio.create_task(coro=task) for task in coros_or_futures
    ]
    try:
        results = await asyncio.gather(*tasks)
    except Exception:
        for task in tasks:
            task.cancel()
        raise
    else:
        return results



from __future__ import annotations

from typing import Dict, Iterable, List, Optional, Tuple, Union

import httpx
from httpx import AsyncClient, Timeout
from typing_extensions import Literal
from volcengine.ApiInfo import ApiInfo
from volcengine.maas.v2.utils import gen_req_id
from volcenginesdkarkruntime import Ark
from volcenginesdkarkruntime._base_client import make_request_options
from volcenginesdkarkruntime._constants import (
    DEFAULT_MAX_RETRIES,
    DEFAULT_TIMEOUT,
)
from volcenginesdkarkruntime._exceptions import ArkAPIError
from volcenginesdkarkruntime._streaming import Stream
from volcenginesdkarkruntime.types.chat import (
    ChatCompletionMessageParam,
    ChatCompletionStreamOptionsParam,
    ChatCompletionToolParam,
    completion_create_params,
)

from ark.core.client import APIGAuthorization, Client
from ark.core.idl.ark_protocol import ArkChatCompletionChunk, ArkChatResponse


class BotClient(Client, Ark):
    def __init__(
        self,
        *,
        base_url: str,
        ak: str | None = None,
        sk: str | None = None,
        jwt_token: str | None = None,
        api_path: str = "/api/v3/bots/chat/completions",
        auto_refresh_apikey: bool = True,
        timeout: float | Timeout | None = DEFAULT_TIMEOUT,
        max_retries: int = DEFAULT_MAX_RETRIES,
        http_client: AsyncClient | None = None,
    ):
        super().__init__(
            ak=ak,
            sk=sk,
            api_key=jwt_token,
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            http_client=http_client,
        )
        self.authorization = None
        if auto_refresh_apikey:
            self.service_id, region = self.parse_apig_host(base_url)
            self.auto_refresh_apikey = auto_refresh_apikey
            self.authorization = APIGAuthorization(ak=ak, sk=sk)
        self.api_info = self.get_api_info(api_path)

    @staticmethod
    def get_api_info(api_path: str) -> Dict[str, ApiInfo]:
        api_info = {
            "chat": ApiInfo("POST", api_path, {}, {}, {}),
        }
        return api_info

    @staticmethod
    def parse_apig_host(base_url: str) -> Tuple[str, str]:
        if not isinstance(base_url, str):
            raise ArkAPIError("base_url is invalid")

        import re

        # base_url format: https://{apig service id}.apigateway-{region}.volceapi.com/
        pattern = r"https://([\w.-]+)\.apigateway-([\w.-])+\.volceapi\.com.*"

        matches = re.match(pattern, base_url)

        if matches and len(matches.group()) >= 2:
            matched_groups = matches.groups()
            service_id, region = matched_groups[0], matched_groups[1]
            return service_id, region
        else:
            raise ArkAPIError("a parameter is missing: service_id")

    def _validate(self) -> None:
        if self.api_key is None and (self.sk is None or self.sk is None):
            raise ArkAPIError("api_key or ak/sk is required")

    def authorize(
        self,
        extra_headers: Optional[Dict[str, str]] = None,
    ) -> Dict[str, str]:
        self._validate()

        if self.authorization:
            self.api_key = self.authorization.get_apikey(service_id=self.service_id)
        if not extra_headers:
            extra_headers = {}
        extra_headers["Authorization"] = f"Bearer {self.api_key}"
        extra_headers["X-Client-Request-Id"] = gen_req_id()
        return extra_headers

    def set_api_key(self, api_key: str) -> None:
        self.api_key = api_key

    def create_chat_completions(
        self,
        *,
        messages: Iterable[ChatCompletionMessageParam],
        model: str,
        frequency_penalty: Optional[float] | None = None,
        function_call: completion_create_params.FunctionCall | None = None,
        functions: Iterable[completion_create_params.Function] | None = None,
        logit_bias: Optional[Dict[str, int]] | None = None,
        logprobs: Optional[bool] | None = None,
        max_tokens: Optional[int] | None = None,
        presence_penalty: Optional[float] | None = None,
        stop: Union[Optional[str], List[str]] | None = None,
        stream: Optional[Literal[False]] | Literal[True] | None = None,
        stream_options: Optional[ChatCompletionStreamOptionsParam] | None = None,
        temperature: Optional[float] | None = None,
        tools: Iterable[ChatCompletionToolParam] | None = None,
        top_logprobs: Optional[int] | None = None,
        top_p: Optional[float] | None = None,
        user: str | None = None,
        metadata: Dict[str, object] | None = None,
        extra_headers: Dict[str, str] | None = None,
        extra_query: Dict[str, object] | None = None,
        extra_body: Dict[str, object] | None = None,
        timeout: float | httpx.Timeout | None = None,
    ) -> Union[ArkChatResponse, Stream[ArkChatCompletionChunk]]:
        api_info = self.api_info.get(
            "chat", ApiInfo("POST", "/api/v3/bots/chat/completions", {}, {}, {})
        )

        extra_headers = self.authorize(extra_headers)

        response = self.chat.completions._post(
            api_info.path,
            body={
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "logprobs": logprobs,
                "max_tokens": max_tokens,
                "presence_penalty": presence_penalty,
                "stop": stop,
                "stream": stream,
                "stream_options": stream_options,
                "temperature": temperature,
                "tools": tools,
                "top_logprobs": top_logprobs,
                "top_p": top_p,
                "user": user,
                "metadata": metadata,
            },
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=ArkChatResponse,
            stream=stream or False,
            stream_cls=Stream[ArkChatCompletionChunk],
        )

        return response



import re
from typing import Iterable, List, get_args

from langchain.schema.output_parser import BaseTransformOutputParser

from ark.core.idl.maas_protocol import SOURCE_TAG, SourceType


class BrowsingIntentMessageChunkOutputParser(BaseTransformOutputParser[bool]):
    """OutputParser that parses BaseMessageChunk into intent of whether to search."""

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Return whether this class is serializable."""
        return True

    @property
    def _type(self) -> str:
        """Return the output parser type for serialization."""
        return "default"

    def parse(self, text: str) -> bool:
        """Returns the input text with no changes."""
        return text == "需要"


class BrowsingGenerationMessageChunkOutputParser(BaseTransformOutputParser[List[str]]):
    """OutputParser that parses BaseMessageChunk into list of str."""

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Return whether this class is serializable."""
        return True

    @property
    def _type(self) -> str:
        """Return the output parser type for serialization."""
        return "default"

    def parse(self, text: str) -> List[str]:
        """Returns the input text with no changes."""
        return text.split("[next]")


class MultiIntentionOutputParser(BaseTransformOutputParser[Iterable[str]]):
    """Parse llm output to bool"""

    source_type: List[SourceType]

    def parse(self, text: str) -> Iterable[str]:
        for source in get_args(SourceType):
            if source not in self.source_type:
                continue
            tag = SOURCE_TAG[source]
            if tag in text:
                yield source


class SearchRewriteOutputParser(BaseTransformOutputParser[List[str]]):
    """Parse llm output to List[str]"""

    def parse(self, text: str) -> List[str]:
        result = re.sub(r"^\d+\.\s-?\s?|\-\s", "", text, flags=re.MULTILINE)
        queries = result.split("\n")
        # filter extreme shot query
        return [q.strip() for q in queries if len(q) >= 2]



import os
from typing import Any, Dict, Optional

from opentelemetry import trace
from opentelemetry.trace.span import Span
from opentelemetry.trace.status import StatusCode

from ark.core.utils.context import get_custom_attributes
from ark.core.utils.types import dump_json_str_truncate

_TRACE_MAX_STRING_LEN = os.environ.get("TRACE_MAX_STRING_LEN", "10000")


def set_trace_attributes(
    span: Span,
    *,
    status_code: trace.StatusCode = StatusCode.UNSET,
    resource_type: str = "",
    resource_id: str = "",
    request_id: str = "",
    client_request_id: str = "",
    account_id: str = "",
    input: Any = None,
    output: Any = None,
    custom_attributes: Optional[Dict[str, Any]] = None,
) -> None:
    span.set_attribute(
        "input", dump_json_str_truncate(input, int(_TRACE_MAX_STRING_LEN))
    )
    span.set_attribute(
        "output", dump_json_str_truncate(output, int(_TRACE_MAX_STRING_LEN))
    )
    span.set_attribute("request_id", request_id)
    span.set_attribute("client_request_id", client_request_id)
    span.set_attribute("resource_type", resource_type)
    span.set_attribute("resource_id", resource_id)
    span.set_attribute("account_id", account_id)
    span.set_status(trace.Status(status_code, ""))

    if custom_attributes is None:
        custom_attributes = get_custom_attributes()
    if custom_attributes:
        for k, v in custom_attributes.items():
            span.set_attribute(k, dump_json_str_truncate(v, int(_TRACE_MAX_STRING_LEN)))



from ark.component.llm.base import (
    BaseChatLanguageModel,
    BaseClassificationLanguageModel,
    BaseEmbeddingLanguageModel,
    BaseTokenizeLanguageModel,
)
from ark.component.llm.llm import (
    chat,
    classification,
    embeddings,
    streaming_chat,
    tokenize,
)

__all__ = [
    "chat",
    "streaming_chat",
    "embeddings",
    "tokenize",
    "classification",
    "BaseChatLanguageModel",
    "BaseEmbeddingLanguageModel",
    "BaseTokenizeLanguageModel",
    "BaseClassificationLanguageModel",
]



import asyncio
import copy
from typing import Any, Callable, Dict, List, Optional

from langchain_core.output_parsers import BaseTransformOutputParser
from pydantic.v1 import Field
from typing_extensions import Literal

from ark.component.output_parser.rag_output import RagRewriteMessageChunkOutputParser
from ark.component.prompts.rag_prompt import (
    CONDENSE_QUESTION_PROMPT_TEMPLATE,
    HYPO_ANSWER_PROMPT_TEMPLATE,
    AugmentPromptTemplate,
    QueryRewritePromptTemplate,
    faq_augment_prompt_template,
    hypo_queries_augment_prompt_template,
    summary_augment_prompt_template,
)
from ark.component.v3.llm.base import BaseChatLanguageModel, BaseChatPromptTemplate
from ark.core.idl.ark_protocol import ArkChatParameters
from ark.core.rag import KnowledgeChunk, KnowledgeSchema
from ark.core.task import task
from ark.core.utils.prompt import format_maas_prompts


def _get_query_rewrite_prompt_template() -> BaseChatPromptTemplate:
    return QueryRewritePromptTemplate(template=CONDENSE_QUESTION_PROMPT_TEMPLATE)


def _get_hypo_answer_rewrite_prompt_template() -> BaseChatPromptTemplate:
    return QueryRewritePromptTemplate(template=HYPO_ANSWER_PROMPT_TEMPLATE)


def _get_default_output_parser() -> BaseTransformOutputParser:
    return RagRewriteMessageChunkOutputParser()


class RewriteGenerator(BaseChatLanguageModel):
    name: Literal["Rewrite", "QueryRewrite", "HypoAnswer"] = "Rewrite"
    template: BaseChatPromptTemplate
    output_parser: BaseTransformOutputParser[List[str]] = Field(
        default_factory=_get_default_output_parser
    )

    def _get_prompt_template(self) -> BaseChatPromptTemplate:
        return self.template

    def _get_output_parser(self) -> BaseTransformOutputParser:
        return self.output_parser

    @task()
    async def arun(self, *args: Any, **kwargs: Any) -> List[str]:
        resp = await super().arun(*args, **kwargs)
        return await self.aparse_output(resp.choices[0].message.content)

    @task()
    def run(self, *args: Any, **kwargs: Any) -> List[str]:
        resp = super().run(*args, **kwargs)
        return self.parse_output(resp.choices[0].message.content)


class QueryRewriteGenerator(RewriteGenerator):
    name: Literal["QueryRewrite"] = "QueryRewrite"
    template: BaseChatPromptTemplate = Field(
        default_factory=_get_query_rewrite_prompt_template
    )


class HypoAnswerGenerator(RewriteGenerator):
    name: Literal["HypoAnswer"] = "HypoAnswer"
    template: BaseChatPromptTemplate = Field(
        default_factory=_get_hypo_answer_rewrite_prompt_template
    )


@task()
async def generate_augment(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    parameters: Optional[ArkChatParameters] = None,
    field_name: str = "text",
    field_prefix: str = "",
    replace_field_with_augment: bool = False,
    template: Optional[AugmentPromptTemplate] = None,
    augment_type: str = "augment",
    augment_parser: Callable[[str], List[str]] = lambda s: [s],
    augment_num_limit: int = 1,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    # run augment
    augment_results = []
    for i in range(0, len(chunks), parallel):
        tasks = []
        for chunk in chunks[i : i + parallel]:
            cur_text = chunk.scalars.get(field_name, None)
            assert cur_text, f"{field_name} not found in chunk.scalar"
            messages = format_maas_prompts(
                template=template,
                chat_messages=[],
                reference=f"{field_prefix}{cur_text}",
            )
            tasks.append(
                BaseChatLanguageModel(
                    endpoint_id=endpoint_id, parameters=parameters, messages=messages
                ).arun()
            )
        augment_results.extend(await asyncio.gather(*tasks))

    # build augment chunks
    augment_chunks = []
    for chunk, augment_resp in zip(chunks, augment_results):
        augment_result = augment_resp.choices[0].message.content
        augments = augment_parser(augment_result)
        for augment in augments[:augment_num_limit]:
            augment_scalars = copy.deepcopy(chunk.scalars)
            doc_id = chunk.scalars["doc_id"]
            chunk_id = chunk.scalars["chunk_id"]
            augment_scalars["chunk_type"] = augment_type
            if replace_field_with_augment:
                augment_scalars.update({field_name: augment, "chunk_len": len(augment)})
            encode_augment = augment[: config.vector_text_len_limit]
            augment_chunks.append(
                KnowledgeChunk(
                    primary_key=(
                        chunk.primary_key[0],
                        "#".join([doc_id, augment_type, str(chunk_id)]),
                    ),
                    vector=(chunk.vector[0], encode_augment),
                    scalars=augment_scalars,
                    knowledge_schema=chunk.knowledge_schema,
                )
            )

    return augment_chunks


@task()
async def generate_summary_augment(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    field_name: str = "text",
    field_prefix: str = "",
    parameters: Optional[Dict[str, Any]] = None,
    augment_num_limit: int = 1,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    return await generate_augment(
        config=config,
        chunks=chunks,
        endpoint_id=endpoint_id,
        field_name=field_name,
        field_prefix=field_prefix,
        replace_field_with_augment=True,
        template=summary_augment_prompt_template,
        parameters=parameters,
        augment_type="summary_augment",
        augment_parser=lambda s: [s] if len(s) > 0 and "无法提取" not in s else [],
        augment_num_limit=augment_num_limit,
        parallel=parallel,
    )


@task()
async def generate_hypo_queries_augment(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    field_name: str = "text",
    field_prefix: str = "",
    parameters: Optional[Dict[str, Any]] = None,
    augment_num_limit: int = 1,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    return await generate_augment(
        config=config,
        chunks=chunks,
        endpoint_id=endpoint_id,
        field_name=field_name,
        field_prefix=field_prefix,
        replace_field_with_augment=False,
        template=hypo_queries_augment_prompt_template,
        parameters=parameters,
        augment_type="hypo_queries_augment",
        augment_parser=lambda s: [s] if len(s) > 0 and "无法提取" not in s else [],
        augment_num_limit=augment_num_limit,
        parallel=parallel,
    )


@task()
async def generate_faq_augment(
    config: KnowledgeSchema,
    chunks: List[KnowledgeChunk],
    endpoint_id: str,
    field_name: str = "text",
    field_prefix: str = "",
    parameters: Optional[Dict[str, Any]] = None,
    augment_num_limit: int = 3,
    parallel: int = 5,
) -> List[KnowledgeChunk]:
    def parse_faq_content(content: str) -> List[str]:
        faqs = []
        for qa_pair in content.split("\n\n"):
            lines = qa_pair.split("\n")
            if len(lines) == 2:
                faqs.append(f"问题：{lines[0]}\n回答：{lines[1]}")
        return faqs

    return await generate_augment(
        config=config,
        chunks=chunks,
        endpoint_id=endpoint_id,
        field_name=field_name,
        field_prefix=field_prefix,
        replace_field_with_augment=True,
        template=faq_augment_prompt_template,
        parameters=parameters,
        augment_type="faq_augment",
        augment_parser=parse_faq_content,
        augment_num_limit=augment_num_limit,
        parallel=parallel,
    )



from __future__ import annotations

import time
from datetime import datetime
from typing import Any, List, Tuple, Union

from jinja2 import Template
from langchain.prompts.chat import BaseChatPromptTemplate
from langchain.schema.messages import BaseMessage, HumanMessage


class BrowsingIntentionChatPromptTemplate(BaseChatPromptTemplate):
    """Browsing Intention prompt template."""

    input_variables: List[str] = ["messages"]
    """List of input variables in template messages. Used for validation."""

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format original messages into a list of messages.

        Args:
            messages: list of original messages.

        Returns:
            Formatted message
        """
        if "messages" not in kwargs:
            raise ValueError("Must provide messages: List[BaseMessage]")
        messages = kwargs.pop("messages")

        questions = [msg.content for msg in messages if msg.type == "human"]
        if len(questions) == 0:
            raise ValueError("No user question found in the request")
        elif len(questions) > 1:
            histories, current = questions[:-1], questions[-1]
        else:
            histories, current = [], questions[0]

        while sum(len(msg) for msg in histories) > 1000:
            histories.pop(0)

        prompt = "\n".join(
            [
                *[f"前文问题:{q}" for q in histories],
                "请判断你是否需要借助搜索引擎来回答下面的问题，请回答需要或者不需要。",
                f"问题：{current}",
            ]
        )

        return [HumanMessage(content=prompt, additional_kwargs=kwargs)]


class BrowsingGenerationChatPromptTemplate(BaseChatPromptTemplate):
    """Browsing Generation prompt template."""

    input_variables: List[str] = ["messages"]
    """List of input variables in template messages. Used for validation."""

    def _gen_meta_info(
            self,
            time_info: Union[datetime, None] = None,
            location_info: Union[Tuple[str, str], None] = None,
    ) -> str:
        city, district = location_info if location_info else ("", "")

        weekdays = [
            "星期一",
            "星期二",
            "星期三",
            "星期四",
            "星期五",
            "星期六",
            "星期日",
        ]  # `tm_wday` 0~6, 6 means Sunday
        if time_info is None:
            t = time.localtime(time.time())
            _time_info = time.strftime("%Y年%m月%d日%H时", t)
            _time_info += weekdays[t.tm_wday]
        else:
            _time_info = time_info.strftime("%Y年%m月%d日%H时")
            _time_info += weekdays[time_info.weekday()]

        meta_info = f"当前时间：{_time_info}"
        if city or district:
            meta_info += f"，当前位置：{city or ''}{district or ''}"
        return meta_info

    def format_messages(self, **kwargs: Any) -> List[BaseMessage]:
        """Format original messages into a list of messages.

        Args:
            messages: list of original messages.
            time_info: datetime, if not provided use current time.
            location_info: tuple[str, str], city and district

        Returns:
            Formatted message
        """
        if "messages" not in kwargs:
            raise ValueError("Must provide messages: List[BaseMessage]")

        messages = kwargs.pop("messages")
        format_messages: List[BaseMessage] = [
            msg.copy(deep=True) for msg in messages if msg.type in {"human", "ai"}
        ]

        if len(format_messages) == 0:
            return format_messages

        # 最多参考几轮历史对话
        max_history_len = kwargs.pop("max_history_len", None)
        if max_history_len is not None:
            format_messages = format_messages[-(max_history_len * 2 + 1):]

        meta_info = self._gen_meta_info(
            time_info=kwargs.pop("time_info", None),
            location_info=kwargs.pop("location_info", None),
        )

        format_messages[-1].content = "\n".join(
            [
                meta_info,
                "如果你希望借助搜索引擎回答下面的问题，你会如何搜索。如果需要搜索多次，请用[next]分割。",
                f"问题：{format_messages[-1].content}",
            ]
        )
        return format_messages


MULTI_INTENTION_TEMPLATE = Template(
    """# 任务描述
请根据「历史记录」和「用户问题」，判断应该以下哪些内容源哪些可以获得有助于回答「用户问题」的资料：
{{sources_desc_text}}

# 任务要求
- 可以选择{{enum_list_text}}，或者选择「无需检索」，不要再进行任何补充描述；
{{task_description}}

# 执行任务
遵循任务要求，判断哪些内容源哪些可以获得有助于回答「用户问题」的资料，只需输出内容源，不用描述原因。

历史记录：
<历史记录开始>
{{chat_history}}
<历史记录结束>

用户问题：
{{question}}

判断结果：
"""
)

DEFAULT_TASK_DESCRIPTION = """- 对于需要补充外部信息才能回答的问题，至少选择一个内容源；
- 天气类、新闻类、实时信息类等时效性问题，或提到[最近]，[今天]，[本周]，[这个月]，[几号]等时间信息，至少选择一个内容源。
- 如果用户问题为闲聊，或问题与内容源无关，或问题不需要额外信息帮助回答，只输出「无需检索」；
-「历史记录」和「用户问题」无关联时，只通过「用户问题」判断，忽略「历史记录」；
"""

QUERY_REWRITE_PROMPT_TEMPLATE = Template(
    """# 角色
你是一个搜索专家，擅长将「用户提问」优化为1~2个适用于搜索引擎的「搜索词」，之后用于搜索有用的资料。

# 任务要求
- 如果用户问题和「历史记录」没有直接上下文关系，改写时忽略历史记录；
- 如果用户提问信息不充分，且「历史记录」有相关上下文，结合「历史记录」进行改写；
- 直接输出「搜索词」，多个搜索词之间用"\n"隔开，输出必须纯文本，不能输出其他任何内容；
- 「搜索词」不超过2个，尽量只输出1个搜索词；
- 如果问题有时效性地域性，结合「当前环境信息」进行改写，将“今天”、“昨天”、“本地” 等相对时间和地理位置使用准确的日期和地理位置添加到搜索词中；
- 如果用户询问的问题有关穿衣建议、户外活动建议或者防护防晒建议，请将「未来天气」也添加到搜索词中；
- 如果用户询问未来天气相关信息，请确保将关键词“未来天气”和用户感兴趣的地理位置添加到搜索词中，地理位置需要是正式的行政区划名，例如“北京未来天气”、“深圳南山区未来天气”等；
- 如果用户想对节日进行日期计算，你需要在「搜索词」中加入合适的年份：
    + 如用户询问某节日已经过去多少天，「搜索词」应有2行，分别加入去年({{last_year}})和今年({{current_year}})的年份
    + 如果用户询问某节日详情（或某节日还剩多少天），「搜索词」应有2行，分别加入今年({{current_year}})和明年({{next_year}})的年份

# 执行任务
遵循任务要求，将「用户提问」优化为1~2个完整独立的利于搜索引擎使用的「搜索词」，「搜索词」不超过100字。

当前环境信息：
{{meta_info}}

历史记录：
<历史记录开始>
{{chat_history}}
<历史记录结束>

用户问题：
{{question}}

优化后的「搜索词」：
"""
)

CUSTOM_SUMMARY_PROMPT = Template(
    """# 历史对话
{{chat_history}}

# 参考资料
{{reference}}

# 当前环境信息
{{meta_info}}

# 任务
{{task_description}}

# 任务执行
遵循任务要求来回答「用户问题」，给出有帮助的回答。

用户问题：
{{question}}

# 你的回答：
"""
)

BROWSING_PARTIAL_TASK_DESCRIPTION = """你要结合最新从各种来源搜索得到的「参考资料」和「当前环境信息」回答「用户问题」。
回答问题时，先查阅与「用户问题」有关联的「参考资料」、「当前环境信息」，结合你自己的知识，汇总信息后，给出对「用户问题」有帮助的回答。

回答要满足以下要求：
1. 使用专业的语言风格回答，使用参考资料时重新组织语言，不得直接复述参考资料；
2. 时效性的问题优先参考以下来源：墨迹天气、搜索引擎；
3. 严肃的问题优先参考以下来源：抖音百科、知识库；
4. 当回答时间、日期、节日的相关问题时，你可以将「当前时间」作为提问时的具体时间，如果「参考资料」没有提供信息，请根据「当前时间」给出一个计算结果
5. 「参考资料」中的国际时间可能有误，要综合「当前时间」考虑并计算
6. 不要在回答中描述参考资料的来源或序号，不能输出“根据参考资料X”、“资料X中提到”、“未获取资料”等说法，不要解释自己；
7. 请牢记下周、本周、上周开始日期都是星期一，例如 “2024年6月13日 星期四”的下周指的是“2024年6月17日 星期一 至 2024年6月23日 星期日”，本周指的是“2024年6月10日 星期一 至 2024年6月16日 星期日”，上周指的是“2024年6月3日 星期一 至 2024年6月9日星期日” ；
"""

BROWSING_STRICT_TASK_DESCRIPTION = """你要结合最新从各种来源搜索得到的「参考资料」和「当前环境信息」回答「用户问题」。
回答问题时，先查阅与「用户问题」有关联的「参考资料」、「当前环境信息」，汇总信息后，给出对「用户问题」有帮助的回答。

回答要满足以下要求：
1. 使用专业的语言风格回答，使用参考资料时重新组织语言，不得直接复述参考资料，回答以参考资料为准；
2. 如果参考资料不能帮助你回答用户问题，请回答“抱歉，这个问题我还不会，尝试告诉我更多信息吧”；
3. 时效性的问题优先参考以下来源：墨迹天气、搜索引擎；
4. 严肃的问题优先参考以下来源：抖音百科、知识库；
5. 当回答时间、日期、节日的相关问题时，你可以将「当前时间」作为提问时的具体时间，如果「参考资料」没有提供信息，请根据「当前时间」给出一个计算结果
6. 「参考资料」中的国际时间可能有误，要综合「当前时间」考虑并计算
7. 不要在回答中描述参考资料的来源或序号，不能输出“根据参考资料X”、“资料X中提到”、“未获取资料”等说法，不要解释自己；
8. 请牢记下周、本周、上周开始日期都是星期一，例如 “2024年6月13日 星期四”的下周指的是“2024年6月17日 星期一 至 2024年6月23日 星期日”，本周指的是“2024年6月10日 星期一 至 2024年6月16日 星期日”，上周指的是“2024年6月3日 星期一 至 2024年6月9日星期日” ；
"""

SOURCE_DESC = {
    "search_engine": "提供互联网全网内容，包含极为丰富全面的各领域信息，但准确度和时效性一般，仅在需要联网信息且其他内容源不满足时选择。",
    "weather": "提供晴雨、温度、湿度、空气质量、风向风力等天气信息",
    "toutiao_article": "今日头条app的中等篇幅图文内容，提供时效性和准确性较高的新闻资讯、娱乐新闻、"
                       "生活技巧、旅行攻略、美食分享、科技资讯、金价、油价、房价、股票等信息，优先选择。",
    "douyin_short_video": "抖音app的短视频内容，提供内容丰富且时效性较高的教程、娱乐、美妆穿搭、唱歌舞蹈、"
                          "纪录生活、旅行见闻、萌娃萌宠等丰富的视频和对应描述文本，优先选择。",
    "douyin_baike": "提供人物、生物、艺术、历史、地理、自然科学、社会科学等相对专业客观的百科知识",
    "xigua_feed_video": "西瓜视频app的中等时长横版视频内容，提供影视解读、游戏解说、科普知识、创意手工视频及对应描述文本。",
    "toutiao_short_content": "今日头条app的社交媒体，提供有个人特色的简短图文内容，"
                             "时效性较高准确性一般，注重与粉丝和博主互动，分享生活趣事，抒发情感。",
}



from ark.component.generator.keywords_generator import KeywordsGenerator
from ark.component.generator.rag_augment_generator import (
    generate_abstract,
    generate_faq_agument,
    generate_hypo_queries_agument,
    generate_summary_agument,
)
from ark.component.generator.rag_rewrite_generator import (
    HypoAnswerGenerator,
    QueryRewriteGenerator,
)

__all__ = [
    "KeywordsGenerator",
    "QueryRewriteGenerator",
    "HypoAnswerGenerator",
    "generate_abstract",
    "generate_summary_agument",
    "generate_hypo_queries_agument",
    "generate_faq_agument",
]



import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

from ark.core.task.task import task
from ark.core.utils.common import Singleton


class Client(Singleton):
    pass


class ClientPool(Singleton):
    _registry: Dict[str, Type[Client]] = {}
    clients: Dict[str, Client] = {}

    def __init__(
        self,
        clients: Dict[str, Tuple[Type[Client], Dict[str, Any]]],
    ) -> None:
        for name, (cls, config) in (clients or dict()).items():
            try:
                self.clients[name] = cls(**config)
            except Exception as e:
                logging.error(f"init client pool failed:{e}")
                continue

    def get_client_names(self) -> List[str]:
        return [name for name in self.clients.keys()]

    def get_client(self, name: str) -> Optional[Client]:
        return self.clients.get(name)

    @classmethod
    def register(
        cls, name: Optional[str] = None
    ) -> Callable[[Type[Client]], Type[Client]]:
        def func(wrapped_cls: Type[Client]) -> Type[Client]:
            if name:
                cls._registry[name] = wrapped_cls
            else:
                cls._registry[wrapped_cls.__name__] = wrapped_cls

            return wrapped_cls

        return func

    @classmethod
    async def async_get_client(cls, name: str, config: Dict[str, Any]) -> Client:
        if name not in cls.clients:
            cls.clients[name] = await cls.async_create_client(name, **config)

        return cls.clients[name]

    @classmethod
    async def async_create_client(cls, name: str, *args: Any, **kwargs: Any) -> Client:
        if name not in cls._registry:
            raise ValueError(f"Unknown client name: {name}")
        if not issubclass(cls._registry[name], Client):
            raise ValueError(f"{name} is not a subclass of client")
        client_cls = cls._registry[name]

        return await client_cls.get_instance_async(*args, **kwargs)


@task()
def get_client_pool(
    clients: Optional[Dict[str, Tuple[Type[Client], Any]]] = None,
) -> ClientPool:
    return ClientPool.get_instance_sync(clients=clients)



from ark.core.assistant.client import AssistantClient
from ark.core.assistant.middleware import LogIdMiddleware
from ark.core.assistant.server import AssistantServer

__all__ = ["AssistantServer", "LogIdMiddleware", "AssistantClient"]



from typing import Any, Dict, Optional

from ark.core.config import ActionConfig


class RagConfig(ActionConfig):
    # rag
    collection_name: str
    project: str = "default"
    retrieve_count: int = 5
    rerank_retrieve_count: Optional[int] = None
    rerank_switch: bool = False
    dense_weight: Optional[float] = 0.5
    show_reference: bool = True
    query_param: Optional[Dict[str, Any]] = None



import hashlib
import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple, Union
from urllib.parse import urlparse

from pydantic import BaseModel, field_validator

ACCEPTED_SCHEMES = {"tos", "http", "https"}


class DocType(Enum):
    PDF = "pdf"
    DOC = "doc"
    DOCX = "docx"
    TXT = "txt"
    MARKDOWN = "md"
    HTML = "html"
    CSV = "csv"
    JSONL = "jonsl"
    UNDEFINED = "undefined"
    PPTX = "pptx"


class KnowledgeDoc(BaseModel):
    """
    This class is used to handle a doc
    """

    local_path: Optional[str] = None
    remote_path: Optional[str] = None
    type: DocType
    encoding: str = "utf-8"
    _doc_id: Optional[str] = None

    doc_name: str = ""
    add_type: str = "url"
    doc_scalars: Optional[List[Dict[str, Any]]] = None

    @property
    def doc_id(self) -> str:
        if self._doc_id:
            return self._doc_id
        if self.remote_path:
            self._doc_id = hashlib.md5(self.remote_path.encode("utf-8")).hexdigest()
        elif self.local_path:
            file_name = self.local_path.split("/")[-1]
            self._doc_id = hashlib.md5(file_name.encode("utf-8")).hexdigest()
        else:
            raise ValueError("local_path and remote_path are both None")
        return self._doc_id
        # return self.doc_scalars['doc_id']

    @field_validator("remote_path")
    @classmethod
    def validate_remote_path(cls, v: Any) -> Optional[str]:
        """check if the remote_path is tos"""
        if v is None:
            return None
        result = urlparse(v)
        if result.scheme not in ACCEPTED_SCHEMES:
            raise ValueError("remote_path must be tos or http or https")
        return v

    @field_validator("local_path")
    @classmethod
    def validate_local_path(cls, v: Any) -> Optional[str]:
        """check if the local_path exists"""
        if v is None:
            return None
        os.path.exists(v)
        return v


class KnowledgeSchema(BaseModel, ABC):
    """
    This class describe Knowledge base index structure
    """

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    primary_key: Dict[str, str]
    vector: Dict[str, Any]
    scalars: List[Dict[str, Any]]

    vector_text_len_limit: int = 500

    @field_validator("primary_key")
    @abstractmethod
    def validate_primary_key(cls, v: Dict[str, str]) -> Dict[str, str]:
        pass

    @field_validator("vector")
    @abstractmethod
    def validate_vector(cls, v: Dict[str, Any]) -> Dict[str, Any]:
        pass

    @field_validator("scalars")
    @abstractmethod
    def validate_scalars(cls, v: List[Dict[str, str]]) -> List[Dict[str, str]]:
        pass

    @abstractmethod
    def get_primary_key_data(self, field: Dict[str, Any]) -> Tuple[str, Any]:
        pass

    @abstractmethod
    def get_vector_data(
        self, field: Dict[str, Any]
    ) -> Tuple[str, Union[List[Any], str]]:
        pass

    @abstractmethod
    def get_scalar_data(self, field: Dict[str, Any]) -> Dict[str, Any]:
        pass


class KnowledgeChunk(BaseModel):
    """
    This class is used to handle a chunk
    """

    primary_key: Tuple[str, Any]
    vector: Tuple[str, Union[List[Any], str]]
    scalars: Dict[str, Any]
    retrieve_score: float = 0.0
    knowledge_schema: Optional[KnowledgeSchema] = None


class KnowledgeBase(BaseModel, ABC):
    """
    This main class
    """

    sid: str
    knowledge_schema: KnowledgeSchema

    @abstractmethod
    def upsert_chunks(self, data: List[KnowledgeChunk]) -> bool:
        """
        Add data to the knowledge base
        """
        pass

    @abstractmethod
    def get_chunk(self, chunk_id: str) -> KnowledgeChunk:
        """
        Get a chunk from the knowledge base
        """
        pass

    @abstractmethod
    def delete_chunks(self, chunk_id: List[str]) -> None:
        """
        Delete chunkfrom the knowledge base
        """
        pass

    @abstractmethod
    def search(
        self, query: Union[str, List[Any]], count: int, **kwargs: Any
    ) -> List[KnowledgeChunk]:
        """
        Search the knowledge base
        """
        pass

    @abstractmethod
    async def aupsert_chunks(self, data: List[KnowledgeChunk]) -> bool:
        """
        Add data to the knowledge base
        """
        pass

    @abstractmethod
    async def aget_chunk(self, chunk_id: str) -> KnowledgeChunk:
        """
        Get a chunk from the knowledge base
        """
        pass

    @abstractmethod
    async def adelete_chunks(self, chunk_id: List[str]) -> None:
        """
        Delete chunkfrom the knowledge base
        """
        pass

    @abstractmethod
    async def asearch(
        self, query: Union[str, List[Any]], count: int, **kwargs: Any
    ) -> List[KnowledgeChunk]:
        """
        Search the knowledge base
        """
        pass



DEFAULT_MODEL_PHASE = "default"
PHASE_INTENTION = "intention"
PHASE_QUERY_REWRITE = "query_rewrite"
PHASE_SEARCH_SUMMARY = "search_summary"
PHASE_SUMMARY = "summary"
PHASE_ORGANIZE_REFERENCE = "organize_reference"

PHASE_SEARCH_INTENTION = "search_intention"
PHASE_SEARCH_INTENTION_BASIC = "search_intention_basic"
PHASE_SEARCH_REWRITE = "search_query_rewrite"
PHASE_SEARCH = "search"
PHASE_RAG_INTENTION = "rag_intention"
PHASE_RAG_REWRITE = "rag_query_rewrite"
PHASE_RAG_RETRIEVE = "rag_retrieve"

CUSTOM_VARIABLE_KEY_TASK_DESCRIPTION = "task_description"
CUSTOM_VARIABLE_KEY_KNOWLEDGE_BASE_DESCRIPTION = "knowledge_base_description"

ACTION_CONFIG_KEY_ACTION_NAME = "action_name"
ACTION_CONFIG_KEY_SOURCE_TYPE = "source_type"
ACTION_NAME_RAG_SNAKE = "rag"
ACTION_NAME_WEBBROWSING_SNAKE = "web_browsing"

DEFAULT_RESULT_MAPPING = {"需要": True, "不需要": False}

# v2 plugins
CALCULATOR_V2_ACTION_NAME = "Calculator"
CALCULATOR_V2_TOOL_NAME = "Calculator"
LINK_READER_V2_ACTION_NAME = "LinkReader"
LINK_READER_V2_TOOL_NAME = "LinkReader"
CONTENT_PLUGIN_ACTION_NAME = "content_plugin"
CONTENT_PLUGIN_TOOL_NAME = "search"



from typing import Any, List, Tuple, Union

from langchain.schema.output_parser import BaseTransformOutputParser
from pydantic.v1 import Field
from typing_extensions import Literal
from volcenginesdkarkruntime.types.completion_usage import CompletionUsage

from ark.component.output_parser.browsing_output import (
    BrowsingGenerationMessageChunkOutputParser,
)
from ark.component.prompts import BrowsingGenerationChatPromptTemplate
from ark.component.v3.llm.base import BaseChatLanguageModel, BaseChatPromptTemplate
from ark.core.task.task import task


class BrowsingQueryRewrite(BaseChatLanguageModel):
    name: Literal["BrowsingQueryRewrite"] = "BrowsingQueryRewrite"
    template: BaseChatPromptTemplate = Field(
        default_factory=BrowsingGenerationChatPromptTemplate
    )
    output_parser: BaseTransformOutputParser[List[str]] = Field(
        default_factory=BrowsingGenerationMessageChunkOutputParser
    )

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @task()
    async def arun(
        self, include_usage: bool = False, *args: Any, **kwargs: Any
    ) -> Union[Tuple[List[str], CompletionUsage], List[str]]:
        resp = await super().arun(*args, **kwargs)
        parsed_content = await self.aparse_output(resp.choices[0].message.content)
        if include_usage:
            return parsed_content, resp.usage
        else:
            return parsed_content



from pydantic.v1 import Field
from typing_extensions import Literal

from ark.component.prompts.custom_prompt import CustomPromptTemplate
from ark.component.prompts.rag_prompt import DEFAULT_SUMMARY_PROMPT
from ark.component.v3.llm.base import BaseChatLanguageModel, BaseChatPromptTemplate


def _default_summary_prompt() -> CustomPromptTemplate:
    return CustomPromptTemplate(
        template=DEFAULT_SUMMARY_PROMPT,
        keep_history_systems=True,
        keep_history_questions=True,
    )


class RagSummary(BaseChatLanguageModel):
    name: Literal["RagSummary"] = "RagSummary"
    template: CustomPromptTemplate = Field(default_factory=_default_summary_prompt)

    def _get_prompt_template(self) -> BaseChatPromptTemplate:
        return self.template



import json
import logging
from typing import Callable, Dict, List, Optional

from volcengine.Credentials import Credentials

from ark.component.reader.base_reader import BaseReader
from ark.core.rag import DocType, KnowledgeChunk, KnowledgeDoc, KnowledgeSchema
from ark.core.task import task
from ark.core.utils.tos import download_tos_file

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)


def read_csv(file_path: str, encoding: str = "utf-8") -> List[Dict]:
    import pandas as pd

    df = pd.read_csv(file_path, encoding=encoding)
    return df.to_dict(orient="records")


def read_jsonl(file_path: str, encoding: str = "utf-8") -> List[Dict]:
    records: List[dict] = []
    err_cnt = 0
    with open(file_path, "r", encoding=encoding) as f:
        for line in f.readlines():
            try:
                r: dict = json.loads(line)
            except Exception:
                err_cnt += 1
            records.append(r)
        if err_cnt > 0:
            LOGGER.warning("json decode error_cnt: %d", err_cnt)
    return records


class StructReader(BaseReader):
    def __init__(
        self,
        config: KnowledgeSchema,
        credentials: Credentials,
        region: str = "cn-beijing",
    ):
        self.config = config
        self.credentials = credentials
        self.region = region

    @task()
    def load(
        self,
        doc: KnowledgeDoc,
        encode_fields: List[str],
        field_map: Optional[Dict[str, str]] = None,
        field_parser: Optional[Dict[str, Callable[[str], str]]] = None,
        cache_path: str = "/tmp/",
        only_save_saclar_fields: bool = True,
    ) -> List[KnowledgeChunk]:
        LOGGER.info("StructReader load started")
        if not doc.local_path:
            doc.local_path = download_tos_file(
                url=str(doc.remote_path),
                credentials=self.credentials,
                cache_path=cache_path,
                region=self.region,
            )
        file_name = doc.local_path.split("/")[-1]
        if doc.type == DocType.CSV:
            records = read_csv(doc.local_path, encoding=doc.encoding)
        elif doc.type == DocType.JSONL:
            records = read_jsonl(doc.local_path, encoding=doc.encoding)
        else:
            raise ValueError(f"Invalid doc type {doc.type}")

        if field_parser:
            for rec in records:
                for k in rec:
                    if k not in field_parser:
                        continue
                    try:
                        parser = field_parser[k]
                        rec[k] = parser(rec[k])
                    except Exception:
                        LOGGER.error("field:%s parse error")

        if field_map:
            new_records = []
            for rec in records:
                new_records.append(
                    {field_map.get(key, key): value for key, value in rec.items()}
                )
            records = new_records

        chunk_type = "raw_chunk"
        chunks = []
        primary_key = self.config.primary_key.get("name", "")
        vector_key = self.config.vector.get("name", "")
        valid_keys = set([data.get("name") for data in self.config.scalars])
        for i, record in enumerate(records):
            valid_dict = {
                "doc_id": doc.doc_id,
                "chunk_id": i,
                "chunk_type": chunk_type,
                "tos_url": doc.remote_path,
                "file_name": file_name,
            }
            if only_save_saclar_fields:
                valid_dict.update({k: v for k, v in record.items() if k in valid_keys})
            else:
                valid_dict.update(record)

            encode_text = ""
            for encode_field in encode_fields:
                if encode_field in valid_dict:
                    encode_text += f"{encode_field}:{valid_dict[encode_field]}\n"

            chunks.append(
                KnowledgeChunk(
                    primary_key=(
                        primary_key,
                        "#".join([doc.doc_id, chunk_type, str(i)]),
                    ),
                    vector=(
                        vector_key,
                        encode_text[: self.config.vector_text_len_limit],
                    ),
                    scalars=valid_dict,
                    knowledge_schema=self.config,
                )
            )
        LOGGER.info("load finished, splitted to %d chunks", len(chunks))
        LOGGER.info("StructReader load finished")
        return chunks



"protocol for GPTEngine"

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field, field_validator
from typing_extensions import Literal

from ark.core.idl.common_protocol import (
    ActionUsage,
    BotUsage,
    ChatMessageContent,
    ChatRole,
    ChatUsage,
    Embedding,
    Error,
    FunctionCall,
    LabelLogprobosValue,
    Parameters,
    Reference,
    Request,
    Response,
    Tool,
)


# maas handler protocol
class ToolCall(BaseModel):
    id: Optional[str] = None
    type: Optional[str] = None
    function: Optional[FunctionCall] = None


class MaasChatMessage(BaseModel):
    role: Optional[ChatRole] = None
    content: Optional[Union[str, List[ChatMessageContent]]] = None
    name: Optional[str] = None
    references: Optional[List[Reference]] = None
    tool_calls: Optional[List[ToolCall]] = None
    tool_call_id: Optional[str] = None


class MaasChatRequest(Request):
    messages: List[MaasChatMessage]
    stream: bool = False
    verbose: bool = False
    crypto_token: Optional[str] = None
    tools: Optional[List[Tool]] = None
    parameters: Optional[Union[Parameters, Dict[str, Any]]] = None
    user: Optional[str] = None
    extra: Dict[str, str] = Field(default_factory=dict)

    def __init__(self, **kwargs: Any):
        super().__init__(**kwargs)

    @field_validator("messages")
    def validate_messages(
        cls, v: Optional[List[MaasChatMessage]]
    ) -> Optional[List[MaasChatMessage]]:
        if not v:
            raise ValueError("Messages are required")

        pre_role = None  # user, assistant, function, system
        for msg in v:
            if not msg.role:
                raise ValueError("Role is required")
            if msg.content is None:
                msg.content = ""

            if msg.role == ChatRole.SYSTEM:
                continue

            if msg.role == ChatRole.FUNCTION:
                assert (
                    pre_role == ChatRole.ASSISTANT
                ), "Function message should follow an assistant message"

            pre_role = msg.role

        return v

    def get_user_info_extra(self) -> Optional[UserInfoExtra]:
        if not self.extra or not self.extra.get("user_info"):
            return None
        return UserInfoExtra.model_validate_json(self.extra.get("user_info", ""))

    def is_emit_intention_signal_extra(self) -> bool:
        return self.extra.get("emit_intention_signal_extra", "false") == "true"


class Logprobs(BaseModel):
    text_offset: Optional[List[int]] = None
    token_logprobs: Optional[List[float]] = None
    tokens: Optional[List[str]] = None
    top_logprobs: Optional[List[Dict[str, Any]]] = None


class MaasChatChoiceLog(BaseModel):
    stage: Optional[str] = None
    input: Optional[str] = None
    content: Optional[str] = None


class MaasChatChoice(BaseModel):
    index: Optional[int] = None
    message: MaasChatMessage
    finish_reason: Optional[str] = None
    logprobs: Optional[Logprobs] = None
    action: Optional[MaasChatChoiceLog] = None
    observation: Optional[MaasChatChoiceLog] = None
    thought: Optional[MaasChatChoiceLog] = None


class MaasChatResponse(Response):
    req_id: Optional[str] = None
    error: Optional[Error] = None
    choices: List[MaasChatChoice] = Field(default_factory=list)
    usage: Optional[Union[ChatUsage, BotUsage]] = None
    extra: Optional[Dict[str, str]] = None

    @staticmethod
    def merge(responses: List[MaasChatResponse]) -> MaasChatResponse:
        merged = responses[0]
        for resp in responses[1:]:
            for i, j in zip(merged.choices, resp.choices):
                if isinstance(i.message.content, str) and isinstance(
                    j.message.content, str
                ):
                    i.message.content += j.message.content
                elif isinstance(i.message.content, list) and isinstance(
                    j.message.content, list
                ):
                    i.message.content.extend(j.message.content)
                else:
                    raise TypeError("no supported merge type")
                i.finish_reason = j.finish_reason
        merged.usage = responses[-1].usage
        merged.extra = responses[-1].extra
        return merged

    def transform_bot_usage(self) -> None:
        if not self.usage:
            return
        if isinstance(self.usage, BotUsage):
            return
        self.usage = BotUsage(model_usage=[self.usage])
        return

    def set_extra(self, key: str, value: str) -> None:
        if not self.extra:
            self.extra = dict()
        self.extra[key] = value


class MaasCertRequest(Request):
    pass


class MaasCertResponse(Response):
    cert: str = ""


class MaasClassificationRequest(Request):
    query: Optional[str] = None
    labels: List[str]


class MaasClassificationResponse(Response):
    req_id: Optional[str] = None
    label: Optional[str] = None
    label_logprobos: Optional[Dict[str, LabelLogprobosValue]] = None
    usage: Optional[ChatUsage] = None
    error: Optional[Error] = None


class MaasTokenizeRequest(Request):
    text: Optional[str] = None


class MaasTokenizeResponse(Response):
    req_id: Optional[str] = None
    total_tokens: Optional[int] = None
    tokens: Optional[List[str]] = None
    token_ids: Optional[List[int]] = None
    offset_mapping: Optional[List[List[int]]] = None
    error: Optional[Error] = None


class MaasEmbeddingsRequest(Request):
    input: Optional[List[str]] = None
    encoding_format: Optional[str] = None
    user: Optional[str] = None


class MaasEmbeddingsResponse(Response):
    req_id: Optional[str] = None
    usage: Optional[ChatUsage] = None
    error: Optional[Error] = None
    object: Optional[str] = None
    data: Optional[List[Embedding]] = None


class UserInfoExtra(BaseModel):
    city: str = ""
    district: str = ""


SourceType = Literal[
    "weather",
    "douyin_baike",
    "toutiao_article",
    "douyin_short_video",
    "search_engine",
    "xigua_feed_video",
    "toutiao_short_content",
]

SOURCE_TAG = {
    "search_engine": "搜索引擎",
    "weather": "墨迹天气",
    "toutiao_article": "头条图文",
    "douyin_short_video": "抖音视频",
    "douyin_baike": "抖音百科",
    "xigua_feed_video": "西瓜视频",
    "toutiao_short_content": "微头条",
}

SearchAction = Literal["WebBrowsingPlus", "WebBrowsingPro", "content_plugin"]


class SearchRequest(Request):
    question: str = ""
    action_name: SearchAction = Field(default="WebBrowsingPlus")
    source_type: Optional[List[SourceType]] = None
    summary_top_k: int = 10
    utm_source: Optional[str] = None
    keywords: Optional[List[str]] = None
    user_info: Optional[UserInfoExtra] = None


class SearchResponse(Response):
    search_results: List[str] = Field(default_factory=list)
    references: List[Reference] = Field(default_factory=list)
    search_usage: Optional[List[ActionUsage]] = None


class KnowledgeBaseResponse(Response):
    texts: List[str] = Field(default_factory=list)
    references: List[Reference] = Field(default_factory=list)



import logging
from typing import Any, List, Optional, Tuple

from langchain.text_splitter import RecursiveCharacterTextSplitter
from volcengine.Credentials import Credentials

from ark.component.reader.base_reader import BaseReader
from ark.core.rag import DocType, KnowledgeChunk, KnowledgeDoc, KnowledgeSchema
from ark.core.task import task
from ark.core.utils.tos import download_tos_file

logging.basicConfig(
    level=logging.INFO, format="[%(asctime)s][%(levelname)s] %(message)s"
)
LOGGER = logging.getLogger(__name__)


def read_pdf(file_path: str) -> Tuple[str, List[str]]:
    try:
        from pypdf import PdfReader
    except ImportError:
        raise ImportError(
            "Could not import pypdf python package. "
            "Please install it with `pip install pypdf`."
        )

    reader = PdfReader(file_path)
    full_text = ""
    pages = []
    for page in reader.pages:
        full_text += page.extract_text()
        pages.append(page.extract_text())
    return full_text, pages


def read_docx(file_path: str) -> Tuple[str, List[Any]]:
    try:
        from docx import Document
    except ImportError:
        raise ImportError(
            "Could not import docx python package. "
            "Please install it with `pip install python-docx`."
        )
    document = Document(file_path)
    full_text = ""
    paragraphs = []
    for paragraph in document.paragraphs:
        full_text += paragraph.text + "\n"
        paragraphs.append(paragraph)
    return full_text, paragraphs


def read_txt(file_path: str, encoding: str = "utf-8") -> Tuple[str, List[str]]:
    lines = []
    with open(file_path, "r", encoding=encoding) as fp:
        lines = [line for line in fp.readlines()]
    full_text = "\n".join(lines)
    return full_text, lines


def read_html(file_path: str, encoding: str = "utf-8") -> Tuple[str, List[str]]:
    import re

    from bs4 import BeautifulSoup

    with open(file_path, "r", encoding=encoding) as fp:
        soup = BeautifulSoup(fp, "html.parser")

    parsed_text = soup.get_text()
    # 处理 html 常见的一些 case
    parsed_text = re.sub(r"[\u3000\u00A0\u2000-\u200B]", " ", parsed_text)
    parsed_text = parsed_text.replace("\r\n", "\n")
    parsed_text = parsed_text.replace(" \n", "\n")
    parsed_text = re.sub(r"\n{3,}", "\n\n", parsed_text)
    return parsed_text, parsed_text.split("\n\n")


class CommonReader(BaseReader):
    """
    Reader for knowledge base
    """

    def __init__(
        self,
        config: KnowledgeSchema,
        credentials: Credentials,
        split_size: Optional[int] = None,
        region: str = "cn-beijing",
    ):
        self.config = config
        self.credentials = credentials
        self.region = region
        self.split_size = config.vector_text_len_limit
        if split_size:
            self.split_size = split_size
        if self.split_size > config.vector_text_len_limit:
            LOGGER.warning(
                "split_size:%d, > vector_text_len_limit:%d",
                split_size,
                config.vector_text_len_limit,
            )

    def _simple_split(self, full_text: str, split_size: int) -> List[str]:
        text_pieces = []
        for i in range(0, len(full_text), split_size):
            text_piece = full_text[i : i + split_size]
            text_pieces.append(text_piece)
        return text_pieces

    def _recursive_split(self, full_text: str, split_size: int) -> List[str]:
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=split_size, chunk_overlap=0
        )
        lc_splited = text_splitter.create_documents([full_text])
        text_pieces = []
        for lc_doc in lc_splited:
            text_pieces.append(lc_doc.page_content)
        return text_pieces

    @task()
    def load(
        self,
        doc: KnowledgeDoc,
        split_method: str = "recursive",
        cache_path: str = "/tmp/",
    ) -> List[KnowledgeChunk]:
        LOGGER.info("CommonReader load started")
        if not doc.local_path:
            doc.local_path = download_tos_file(
                url=str(doc.remote_path),
                credentials=self.credentials,
                cache_path=cache_path,
                region=self.region,
            )
        file_name = doc.local_path.split("/")[-1]
        if doc.type == DocType.PDF:
            full_text, _ = read_pdf(doc.local_path)
        elif doc.type == DocType.TXT or doc.type == DocType.MARKDOWN:
            full_text, _ = read_txt(doc.local_path, encoding="utf-8")
        elif doc.type == DocType.DOC or doc.type == DocType.DOCX:
            full_text, _ = read_docx(doc.local_path)
        elif doc.type == DocType.HTML:
            full_text, _ = read_html(doc.local_path)
        else:
            raise ValueError(f"Invalid doc type {doc.type}")
        primary_key = self.config.primary_key.get("name", "")
        vector_key = self.config.vector.get("name", "")
        full_text_chunk = KnowledgeChunk(
            primary_key=(primary_key, "#".join([doc.doc_id, "full_text", str(-1)])),
            vector=(
                vector_key,
                file_name + "\n" + full_text[: self.split_size - len(file_name) - 1],
            ),
            scalars={
                "doc_id": doc.doc_id,
                "chunk_id": -1,
                "chunk_type": "full_text",
                "text": full_text,
                "tos_url": doc.remote_path,
                "file_name": file_name,
                "chunk_len": len(full_text),
                "full_text_len": len(full_text),
            },
            knowledge_schema=self.config,
        )
        if split_method == "simple":
            text_pieces = self._simple_split(full_text, self.split_size)
        elif split_method == "recursive":
            text_pieces = self._recursive_split(full_text, self.split_size)
        chunk_type = "raw_chunk"
        chunks = [
            KnowledgeChunk(
                primary_key=(primary_key, "#".join([doc.doc_id, chunk_type, str(i)])),
                vector=(vector_key, text_piece[: self.split_size]),
                scalars={
                    "doc_id": doc.doc_id,
                    "chunk_id": i,
                    "chunk_type": chunk_type,
                    "text": text_piece,
                    "tos_url": doc.remote_path,
                    "file_name": file_name,
                    "chunk_len": len(text_piece),
                    "full_text_len": len(full_text),
                },
                knowledge_schema=self.config,
            )
            for i, text_piece in enumerate(text_pieces)
        ]
        LOGGER.info("load finished, splitted to %d chunks", len(chunks))
        LOGGER.info("CommonReader load finished")
        return [full_text_chunk] + chunks



from ark.component.plugins.plugin import Plugin
from ark.component.plugins.search_intention import SearchIntention
from ark.component.plugins.search_summary import SearchSummary

__all__ = ["Plugin", "SearchIntention", "SearchSummary"]



import json
import logging
import re
from enum import Enum
from typing import (
    Any,
    AsyncGenerator,
    AsyncIterable,
    Generator,
    Iterable,
    List,
    Optional,
    Type,
    Union,
)

import fastapi
from pydantic import BaseModel, ValidationError
from volcenginesdkarkruntime.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
)

from ark.core.idl.ark_protocol import (
    ArkMessage,
    ChatCompletionMessageToolCallParam,
    Function,
)
from ark.core.idl.common_protocol import ChatMessage, RequestType
from ark.core.idl.maas_protocol import MaasChatMessage
from ark.core.utils.errorsv3 import InvalidParameter, MissingParameter

_MAX_DEPTH = 10


def parse_pydantic_error(
    e: ValidationError,
) -> Union[MissingParameter, InvalidParameter]:
    try:
        err = e.errors()[-1]
        err_type, err_loc = err.get("type", ""), err.get("loc", ("",))
        if err_type == "missing":
            return MissingParameter(str(err_loc[-1]))
        else:
            return InvalidParameter(str(err_loc[-1]))
    except Exception as e:
        logging.error(f"parse pydantic error:{e}")
        return InvalidParameter("request")


def dump_json_str(obj: Any) -> str:
    return json.dumps(dump_json(obj), ensure_ascii=False, default=lambda x: str(x))


def dump_json(obj: Any) -> Any:
    if isinstance(obj, dict):
        return {k: dump_json(v) for k, v in obj.items()}
    elif isinstance(obj, (tuple, list)):
        return [dump_json(v) for v in obj]
    elif isinstance(obj, BaseModel):
        return obj.model_dump(exclude_unset=True, exclude_none=True)
    elif isinstance(obj, (AsyncGenerator, Generator, AsyncIterable)):
        return str(obj)
    else:
        return obj


# Only for trace
# Truncate all strings to prevent excessively long input/output (e.g., base64 images)
def dump_json_str_truncate(obj: Any, string_length_limit: int) -> str:
    return json.dumps(
        dump_json_truncate(obj, string_length_limit),
        ensure_ascii=False,
        default=lambda x: str(x),
    )


def dump_json_truncate(obj: Any, string_length_limit: int, depth: int = 0) -> Any:
    if depth > _MAX_DEPTH:  # for safety
        return "max recursion depth exceeded"
    if isinstance(obj, dict):
        result_dict = {}
        for k, v in obj.items():
            value = dump_json_truncate(v, string_length_limit, depth + 1)
            if value is not None:
                result_dict[k] = value
        return result_dict
    elif isinstance(obj, Enum):
        return obj.value
    elif isinstance(obj, (AsyncGenerator, Generator, AsyncIterable)):
        return str(obj)
    elif isinstance(obj, (list, tuple)):
        return [
            dump_json_truncate(item, string_length_limit, depth + 1) for item in obj
        ]
    elif isinstance(obj, str):
        return obj[:string_length_limit]
    elif isinstance(obj, BaseModel) and hasattr(obj, "__dict__"):
        result_dict = {}
        for k, v in obj.__dict__.items():
            value = dump_json_truncate(v, string_length_limit, depth + 1)
            if value is not None:
                result_dict[k] = value
        return result_dict
    else:
        return obj


async def load_request(
    http_request: fastapi.Request,
    req_cls: Type[RequestType],
) -> RequestType:
    if "content-type" not in http_request.headers:
        raise InvalidParameter("Invalid request: missing content-type")
    content_type, body = (
        http_request.headers.get("content-type", ""),
        await http_request.body(),
    )
    media_type = content_type.split(";")[0].strip()
    try:
        if media_type == "application/json":
            return req_cls.model_validate_json(body)
    except ValidationError as e:
        raise parse_pydantic_error(e)
    raise InvalidParameter(f"Invalid request: invalid content-type={content_type}")


def transform_response(response: Any) -> str:
    if isinstance(response, str):
        return response
    elif isinstance(response, dict):
        return json.dumps(response)
    elif isinstance(response, BaseModel):
        return response.model_dump_json(exclude_none=True, exclude_unset=True)
    else:
        return dump_json_str(response)


def convert_messages(
    messages: Optional[List[MaasChatMessage]] = None,
) -> List[ChatMessage]:
    if messages is None:
        raise ValueError("messages are required")

    result: List[ChatMessage] = []
    for message in messages:
        if not isinstance(message.content, str):
            continue
        result.append(ChatMessage(role=message.role, content=message.content))

    return result


def convert_response_message(
    response_message: Union[ChatCompletionChunk, ChatCompletion],
) -> ArkMessage:
    return ArkMessage(
        role=response_message.role,
        content=response_message.content,
        tool_calls=[
            ChatCompletionMessageToolCallParam(
                id=tool_call.id,
                type=tool_call.type,
                function=Function(**tool_call.function.__dict__),
            )
            for tool_call in response_message.tool_calls
        ]
        if response_message.tool_calls
        else None,
    )


def snake_merge(merge_lists: Iterable[List[Any]]) -> List[Any]:
    result = []
    max_length = max(
        len(merge_list) if merge_list is not None else 0 for merge_list in merge_lists
    )
    for i in range(max_length):
        for merge_list in merge_lists:
            if merge_list is None:
                continue
            if i < len(merge_list):
                result.append(merge_list[i])
    return result


def camel_to_snake_case(obj: Any, ignore_value: bool = False) -> Any:
    # Use regularized expression to convert camelCase to snake_case
    # Add '_' before Upper case letters and change all letters to lowercase

    if isinstance(obj, BaseModel):
        return camel_to_snake_case(obj.dict(exclude_unset=True))
    elif isinstance(obj, dict):
        return {
            camel_to_snake_case(k): v
            if ignore_value and isinstance(v, str)
            else camel_to_snake_case(v)
            for k, v in obj.items()
        }
    elif isinstance(obj, (tuple, list)):
        return [camel_to_snake_case(v) for v in obj]
    elif isinstance(obj, str):
        return re.sub(r"(?<!^)(?=[A-Z])", "_", obj).lower()
    else:
        return obj



import asyncio
import logging
import time
from dataclasses import dataclass, field
from functools import partial
from typing import Awaitable, Callable

import anyio
from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from ark.core.telemetry.logger import DEBUG, INFO, WARN
from ark.core.utils.context import (
    get_client_reqid,
    get_reqid,
    get_start_time,
    set_client_reqid,
    set_headers,
    set_reqid,
    set_start_time,
)
from ark.core.utils.logid import gen_log_id


@dataclass
class LogIdMiddleware:
    app: "ASGIApp"
    client_header_name: str = "x-client-request-id"
    header_name: str = "x-request-id"

    generator: Callable[[], str] = field(default=lambda: gen_log_id())

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Load log ID from headers if present. Generate one otherwise.
        And put it into context.
        """
        set_start_time(time.perf_counter())
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        headers = MutableHeaders(scope=scope)
        headers[self.header_name] = headers.get(
            "X-Faas-Request-Id", headers.get(self.header_name, self.generator())
        )
        headers[self.client_header_name] = headers.get(
            self.client_header_name.lower(), headers[self.header_name]
        )

        set_client_reqid(headers[self.client_header_name])
        set_reqid(headers[self.header_name])
        set_headers(headers)

        async def handle_outgoing_request(message: "Message") -> None:
            if message["type"] == "http.response.start" and get_reqid():
                headers = MutableHeaders(scope=message)
                headers.append(self.header_name, get_reqid())
                headers.append(self.client_header_name, get_client_reqid())

            await send(message)
            logging.debug(
                f"[{get_reqid()}] out app cost={time.perf_counter() - get_start_time()}"
            )

        await self.app(scope, receive, handle_outgoing_request)
        return


@dataclass
class ListenDisconnectionMiddleware:
    app: "ASGIApp"

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Cancel the request if the client disconnected.
        """
        INFO("listen disconnection middleware called.")

        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        request_received, response_sent = asyncio.Event(), asyncio.Event()

        async def _receive() -> Message:
            message = await receive()
            DEBUG(f"receive message={message}")
            if message["type"] == "http.request" and not message.get(
                "more_body", False
            ):
                request_received.set()
            return message

        async def _send(message: Message) -> None:
            DEBUG(f"send message={message}")
            if message["type"] == "http.response.body" and not message.get(
                "more_body", False
            ):
                response_sent.set()
            await send(message)

        async def listen_for_disconnect() -> None:
            await request_received.wait()
            while True:
                message = await receive()
                if message["type"] == "http.disconnect":
                    if not response_sent.is_set():
                        WARN("request canceled before response finished.")
                    break

        async with anyio.create_task_group() as task_group:

            async def wrap(func: Callable[[], Awaitable[None]]) -> None:
                await func()
                task_group.cancel_scope.cancel()

            task_group.start_soon(wrap, partial(self.app, scope, _receive, _send))
            await wrap(listen_for_disconnect)

        return



from ark.component.v3.llm.rag_rewrite.rag_rewrite import (
    HypoAnswerGenerator,
    QueryRewriteGenerator,
    RewriteGenerator,
    generate_augment,
    generate_faq_augment,
    generate_hypo_queries_augment,
    generate_summary_augment,
)

__all__ = [
    "RewriteGenerator",
    "QueryRewriteGenerator",
    "HypoAnswerGenerator",
    "generate_augment",
    "generate_summary_augment",
    "generate_hypo_queries_augment",
    "generate_faq_augment",
]



from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Union

from langchain.prompts.chat import BaseChatPromptTemplate
from langchain.schema.output_parser import BaseTransformOutputParser
from pydantic.v1 import BaseModel, Field

from ark.component.output_parser.browsing_output import (
    BrowsingGenerationMessageChunkOutputParser,
)
from ark.component.prompts import BrowsingGenerationChatPromptTemplate
from ark.core._api.deprecation import deprecated
from ark.core.client import ArkClient, Client, get_client_pool
from ark.core.idl.common_protocol import Parameters
from ark.core.idl.maas_protocol import (
    MaasChatMessage,
    MaasChatRequest,
    MaasChatResponse,
)
from ark.core.task import task
from ark.core.utils.prompt import format_maas_prompts


def _get_ark_client() -> Optional[Client]:
    client_pool = get_client_pool()
    client = client_pool.get_client("chat")
    if not client:
        client = ArkClient()
    return client


def _get_default_prompt_template() -> BaseChatPromptTemplate:
    return BrowsingGenerationChatPromptTemplate()


def _get_default_output_parser() -> BaseTransformOutputParser:
    return BrowsingGenerationMessageChunkOutputParser()


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.llm.BrowsingQueryRewrite",
)
class KeywordsGenerator(BaseModel):
    endpoint_id: str
    messages: List[MaasChatMessage]
    parameters: Optional[Union[Parameters, Dict[str, Any]]] = None
    options: Optional[Dict[str, Any]] = None
    client: ArkClient = Field(default_factory=_get_ark_client)
    template: BaseChatPromptTemplate = Field(
        default_factory=_get_default_prompt_template
    )
    output_parser: BaseTransformOutputParser[List[str]] = Field(
        default_factory=_get_default_output_parser
    )

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def generate_prompts(
        self, messages: List[MaasChatMessage], **kwargs: Any
    ) -> List[MaasChatMessage]:
        return format_maas_prompts(self.template, messages, **kwargs)

    def parse_output(self, text: str) -> List[str]:
        return self.output_parser.parse(text)

    async def aparse_output(self, output: str) -> List[str]:
        return await self.output_parser.aparse(output)

    @task()
    async def _arun(
        self,
        endpoint_id: str,
        request: MaasChatRequest,
    ) -> Union[AsyncIterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return await self.client.async_stream_chat(
                endpoint_id=endpoint_id, req=request
            )
        else:
            return await self.client.async_chat(endpoint_id=endpoint_id, req=request)

    @task()
    def _run(
        self, endpoint_id: str, request: MaasChatRequest
    ) -> Union[Iterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return self.client.stream_chat(endpoint_id=endpoint_id, req=request)
        else:
            return self.client.chat(endpoint_id=endpoint_id, req=request.model_dump())

    async def arun(self, **kwargs: Any) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
        )

        return await self._arun(self.endpoint_id, request)

    async def astream(self, **kwargs: Any) -> AsyncIterator[MaasChatResponse]:
        request = MaasChatRequest(
            stream=True,
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
        )

        async for resp in await self._arun(self.endpoint_id, request):
            yield resp

    def run(self, **kwargs: Any) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
        )
        return self._run(self.endpoint_id, request)

    def stream(self, **kwargs: Any) -> Iterator[MaasChatResponse]:
        request = MaasChatRequest(
            stream=True,
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
        )

        for resp in self._run(self.endpoint_id, request):
            yield resp



import inspect
from inspect import Signature
from typing import AsyncIterable, Callable, Dict, Type

from ark.core.idl.ark_protocol import (
    ArkChatRequest,
)
from ark.core.idl.common_protocol import Request, RequestType, ResponseType
from ark.core.idl.maas_protocol import MaasChatRequest, MaasChatResponse
from ark.core.runtime.asyncio import (
    AsyncRunner,
    ChatAsyncRunner,
    ChatV3AsyncRunner,
    CustomAsyncRunner,
)
from ark.core.task.task import task


@task()
def get_runner(
    runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
) -> AsyncRunner:
    signature = inspect.signature(runnable_func)
    request_cls: Type[RequestType] = get_request_cls(signature)
    response_cls: Type[ResponseType] = get_response_cls(signature)

    if issubclass(request_cls, MaasChatRequest) and issubclass(
        response_cls, MaasChatResponse
    ):
        return ChatAsyncRunner(runnable_func)  # type: ignore
    elif issubclass(request_cls, ArkChatRequest):
        return ChatV3AsyncRunner(runnable_func)  # type: ignore
    else:
        return CustomAsyncRunner(response_cls, runnable_func)


@task()
def get_endpoint_config(
    endpoint_path: str,
    runnable_func: Callable[[RequestType], AsyncIterable[ResponseType]],
) -> Dict[str, Type[RequestType]]:
    signature = inspect.signature(runnable_func)
    return {endpoint_path: get_request_cls(signature)}


@task()
def get_request_cls(signature: Signature) -> Type[RequestType]:
    parameters = signature.parameters.values()

    for param in parameters:
        annotation = param.annotation
        assert issubclass(annotation, Request), TypeError(
            "function request should be subclass of request"
        )
        return annotation

    raise Exception("should not reach here")


@task()
def get_response_cls(signature: Signature) -> Type[ResponseType]:
    return_cls = signature.return_annotation
    assert hasattr(return_cls, "__origin__") and issubclass(
        return_cls.__origin__, AsyncIterable
    ), TypeError("function response should be AsyncIterable")
    assert return_cls.__args__ and len(return_cls.__args__) == 1, TypeError(
        "function should return one value"
    )

    # skip response class check
    response_cls = return_cls.__args__[0]
    return response_cls



import functools
import inspect
import warnings
from typing import Any, Callable, Type, TypeVar, Union, cast


class ArkDeprecationWarning(DeprecationWarning):
    """A class for issuing deprecation warnings for Ark users."""


class ArkPendingDeprecationWarning(PendingDeprecationWarning):
    """A class for issuing deprecation warnings for Ark users."""


T = TypeVar("T", bound=Union[Type, Callable[..., Any]])


def warn_deprecated(
    since: str,
    *,
    message: str = "",
    name: str = "",
    alternative: str = "",
    alternative_import: str = "",
    pending: bool = False,
    obj_type: str = "",
    addendum: str = "",
    removal: str = "",
) -> None:
    """Display a standardized deprecation.

    Arguments:
        since : str
            The release at which this API became deprecated.
        message : str, optional
            Override the default deprecation message. The %(since)s,
            %(name)s, %(alternative)s, %(obj_type)s, %(addendum)s,
            and %(removal)s format specifiers will be replaced by the
            values of the respective arguments passed to this function.
        name : str, optional
            The name of the deprecated object.
        alternative : str, optional
            An alternative API that the user may use in place of the
            deprecated API. The deprecation warning will tell the user
            about this alternative if provided.
        pending : bool, optional
            If True, uses a PendingDeprecationWarning instead of a
            DeprecationWarning. Cannot be used together with removal.
        obj_type : str, optional
            The object type being deprecated.
        addendum : str, optional
            Additional text appended directly to the final message.
        removal : str, optional
            The expected removal version. With the default (an empty
            string), a removal version is automatically computed from
            since. Set to other Falsy values to not schedule a removal
            date. Cannot be used together with pending.
    """
    if pending and removal:
        warnings.warn("A pending deprecation cannot have a scheduled removal")
    if alternative and alternative_import:
        warnings.warn("Cannot specify both alternative and alternative_import")

    if not pending:
        if not removal:
            removal = f"in {removal}" if removal else "within ?? minor releases"
            warnings.warn(
                f"Need to determine which default deprecation schedule to use. "
                f"{removal}"
            )
        else:
            removal = f"in {removal}"

    if not message:
        message = ""
        package = name.split(".")[0].replace("_", "-") if "." in name else "ark"

        if obj_type:
            message += f"The {obj_type} `{name}`"
        else:
            message += f"`{name}`"

        if pending:
            message += " will be deprecated in a future version"
        else:
            message += f" was deprecated in {package} {since}"

            if removal:
                message += f" and will be removed {removal}"

        if alternative_import:
            alt_package = alternative_import.split(".")[0].replace("_", "-")
            if alt_package == package:
                message += f". Use {alternative_import} instead."
            else:
                split_import = alternative_import.rsplit(".", 1)
                if len(split_import) > 1:
                    alt_module, alt_name = split_import
                    message += (
                        f". An updated version of the {obj_type} exists in the "
                        f"{alt_package} package and should be used instead."
                        f"`from {alt_module} import {alt_name}`."
                    )
                elif len(split_import) == 1:
                    message += (
                        f". An updated version of the {obj_type} exists in the "
                        f"{alt_package} package and should be used instead."
                        f"`import {split_import[0]}`."
                    )
        elif alternative:
            message += f". Use {alternative} instead."

        if addendum:
            message += f" {addendum}"

    warning_cls = ArkPendingDeprecationWarning if pending else ArkDeprecationWarning
    warning = warning_cls(message)
    warnings.warn(warning, category=ArkDeprecationWarning, stacklevel=2)


def deprecated(
    since: str,
    *,
    message: str = "",
    name: str = "",
    alternative: str = "",
    alternative_import: str = "",
    pending: bool = False,
    obj_type: str = "",
    addendum: str = "",
    removal: str = "",
) -> Callable[[T], T]:
    """Decorator to mark a function, a class, or a property as deprecated.

    When deprecating a classmethod, a staticmethod, or a property, the
    ``@deprecated`` decorator should go *under* ``@classmethod`` and
    ``@staticmethod`` (i.e., `deprecated` should directly decorate the
    underlying callable), but *over* ``@property``.

    When deprecating a class ``C`` intended to be used as a base class in a
    multiple inheritance hierarchy, ``C`` *must* define an ``__init__`` method
    (if ``C`` instead inherited its ``__init__`` from its own base class, then
    ``@deprecated`` would mess up ``__init__`` inheritance when installing its
    own (deprecation-emitting) ``C.__init__``).

    Parameters are the same as for `warn_deprecated`, except that *obj_type*
    defaults to 'class' if decorating a class, 'attribute' if decorating a
    property, and 'function' otherwise.

    Arguments:
        since : str
            The release at which this API became deprecated.
        message : str, optional
            Override the default deprecation message. The %(since)s,
            %(name)s, %(alternative)s, %(obj_type)s, %(addendum)s,
            and %(removal)s format specifiers will be replaced by the
            values of the respective arguments passed to this function.
        name : str, optional
            The name of the deprecated object.
        alternative : str, optional
            An alternative API that the user may use in place of the
            deprecated API. The deprecation warning will tell the user
            about this alternative if provided.
        pending : bool, optional
            If True, uses a PendingDeprecationWarning instead of a
            DeprecationWarning. Cannot be used together with removal.
        obj_type : str, optional
            The object type being deprecated.
        addendum : str, optional
            Additional text appended directly to the final message.
        removal : str, optional
            The expected removal version. With the default (an empty
            string), a removal version is automatically computed from
            since. Set to other Falsy values to not schedule a removal
            date. Cannot be used together with pending.

    Examples
    --------

        .. code-block:: python

            @deprecated('1.4.0')
            def the_function_to_deprecate():
                pass
    """

    def deprecate(
        obj: T,
        *,
        _obj_type: str = obj_type,
        _name: str = name,
        _message: str = message,
        _alternative: str = alternative,
        _alternative_import: str = alternative_import,
        _pending: bool = pending,
        _addendum: str = addendum,
    ) -> T:
        """Implementation of the decorator returned by `deprecated`."""

        def emit_warning() -> None:
            """Emit the warning."""
            warn_deprecated(
                since,
                message=_message,
                name=_name,
                alternative=_alternative,
                alternative_import=_alternative_import,
                pending=_pending,
                obj_type=_obj_type,
                addendum=_addendum,
                removal=removal,
            )

        warned = False

        def warning_emitting_wrapper(*args: Any, **kwargs: Any) -> Any:
            """Wrapper for the original wrapped callable that emits a warning.

            Args:
                *args: The positional arguments to the function.
                **kwargs: The keyword arguments to the function.

            Returns:
                The return value of the function being wrapped.
            """
            nonlocal warned
            if not warned:
                warned = True
                emit_warning()
            return wrapped(*args, **kwargs)

        async def awarning_emitting_wrapper(*args: Any, **kwargs: Any) -> Any:
            """Same as warning_emitting_wrapper, but for async functions."""

            nonlocal warned
            if not warned:
                warned = True
                emit_warning()
            return await wrapped(*args, **kwargs)

        if isinstance(obj, type):
            if not _obj_type:
                _obj_type = "class"
            wrapped = obj.__init__  # type: ignore
            _name = _name or (
                f"{obj.__module__}.{obj.__name__}" if obj.__module__ else obj.__name__
            )
            old_doc = obj.__doc__

            def finalize(_: Any, new_doc: str) -> T:
                """Finalize the deprecation of a class."""
                try:
                    obj.__doc__ = new_doc
                except AttributeError:  # Can't set on some extension objects.
                    pass

                def warn_if_direct_instance(
                    self: Any, *args: Any, **kwargs: Any
                ) -> Any:
                    """Warn that the class is in beta."""
                    nonlocal warned
                    if not warned and type(self) is obj:
                        warned = True
                        emit_warning()
                    return wrapped(self, *args, **kwargs)

                obj.__init__ = functools.wraps(obj.__init__)(  # type: ignore[misc]
                    warn_if_direct_instance
                )
                return cast(T, obj)

        elif isinstance(obj, property):
            if not _obj_type:
                _obj_type = "attribute"
            wrapped = None
            _name = _name or obj.fget.__name__
            old_doc = obj.__doc__

            class _deprecated_property(type(obj)):  # type: ignore
                """A deprecated property."""

                def __get__(self, instance, owner=None):  # type: ignore
                    if instance is not None or owner is not None:
                        emit_warning()
                    return super().__get__(instance, owner)

                def __set__(self, instance, value):  # type: ignore
                    if instance is not None:
                        emit_warning()
                    return super().__set__(instance, value)

                def __delete__(self, instance):  # type: ignore
                    if instance is not None:
                        emit_warning()
                    return super().__delete__(instance)

                def __set_name__(self, owner, set_name):  # type: ignore
                    nonlocal _name
                    if _name == "<lambda>":
                        _name = set_name

            def finalize(_: Any, new_doc: str) -> Any:  # type: ignore
                """Finalize the property."""
                return _deprecated_property(
                    fget=obj.fget, fset=obj.fset, fdel=obj.fdel, doc=new_doc
                )

        else:
            if not _obj_type:
                _obj_type = "function"
            wrapped = obj
            _name = _name or obj.__name__  # type: ignore
            old_doc = wrapped.__doc__

            def finalize(  # type: ignore
                wrapper: Callable[..., Any], new_doc: str
            ) -> T:
                """Wrap the wrapped function using the wrapper and update the docstring.

                Args:
                    wrapper: The wrapper function.
                    new_doc: The new docstring.

                Returns:
                    The wrapped function.
                """
                wrapper = functools.wraps(wrapped)(wrapper)
                wrapper.__doc__ = new_doc
                return cast(T, wrapper)

        old_doc = inspect.cleandoc(old_doc or "").strip("\n")

        # old_doc can be None
        if not old_doc:
            old_doc = ""

        # Modify the docstring to include a deprecation notice.
        notes_header = "\nNotes\n-----"
        components = [
            message,
            f"Use {alternative} instead." if alternative else "",
            addendum,
        ]
        details = " ".join([component.strip() for component in components if component])
        package = _name.split(".")[0].replace("_", "-") if "." in _name else None
        since_str = f"{package}=={since}" if package else since
        new_doc = (
            f"[*Deprecated*] {old_doc}\n"
            f"{notes_header if notes_header not in old_doc else ''}\n"
            f".. deprecated:: {since_str}\n"
            f"   {details}"
        )

        if inspect.iscoroutinefunction(obj):
            finalized = finalize(awarning_emitting_wrapper, new_doc)
        else:
            finalized = finalize(warning_emitting_wrapper, new_doc)
        return cast(T, finalized)

    return deprecate



from typing import Any, AsyncIterator, Dict, Iterator, List, Optional, Union

from langchain_core.output_parsers import BaseTransformOutputParser
from langchain_core.prompts import BaseChatPromptTemplate

from ark.core.idl.common_protocol import Embedding, Parameters
from ark.core.idl.maas_protocol import (
    MaasChatMessage,
    MaasChatRequest,
    MaasChatResponse,
    MaasClassificationRequest,
    MaasClassificationResponse,
    MaasEmbeddingsRequest,
    MaasEmbeddingsResponse,
    MaasTokenizeRequest,
    MaasTokenizeResponse,
)
from ark.core.llm import BaseLanguageModel
from ark.core.task import task
from ark.core.utils.prompt import format_maas_prompts


class BaseChatLanguageModel(BaseLanguageModel[MaasChatResponse]):
    messages: List[MaasChatMessage]
    parameters: Optional[Union[Parameters, Dict[str, Any]]] = None
    template: Optional[BaseChatPromptTemplate] = None
    output_parser: Optional[BaseTransformOutputParser] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def generate_prompts(
        self, messages: List[MaasChatMessage], **kwargs: Any
    ) -> List[MaasChatMessage]:
        if not self.template:
            return messages

        return format_maas_prompts(self.template, messages, **kwargs)

    @task()
    def parse_output(self, text: str) -> Any:
        if not self.output_parser:
            return text

        return self.output_parser.parse(text)

    @task()
    async def aparse_output(self, text: str) -> Any:
        if not self.output_parser:
            return text

        return await self.output_parser.aparse(text)

    @task()
    async def _arun(
        self,
        endpoint_id: str,
        request: MaasChatRequest,
    ) -> Union[AsyncIterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return await self.client.async_stream_chat(
                endpoint_id=endpoint_id, req=request
            )
        else:
            return await self.client.async_chat(endpoint_id=endpoint_id, req=request)

    @task()
    def _run(
        self, endpoint_id: str, request: MaasChatRequest
    ) -> Union[Iterator[MaasChatResponse], MaasChatResponse]:
        if request.stream:
            return self.client.stream_chat(endpoint_id=endpoint_id, req=request)
        else:
            return self.client.chat(endpoint_id=endpoint_id, req=request.model_dump())

    async def arun(self, *args: Any, **kwargs: Any) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
            **kwargs,
        )

        return await self._arun(self.endpoint_id, request)

    async def astream(
        self, *args: Any, **kwargs: Any
    ) -> AsyncIterator[MaasChatResponse]:
        request = MaasChatRequest(
            stream=True,
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
            **kwargs,
        )

        async for resp in await self._arun(self.endpoint_id, request):
            yield resp

    def run(self, *args: Any, **kwargs: Any) -> MaasChatResponse:
        request = MaasChatRequest(
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
            **kwargs,
        )
        return self._run(self.endpoint_id, request)

    def stream(self, **kwargs: Any) -> Iterator[MaasChatResponse]:
        request = MaasChatRequest(
            stream=True,
            messages=self.generate_prompts(self.messages, **kwargs),
            parameters=self.parameters,
            **kwargs,
        )

        for resp in self._run(self.endpoint_id, request):
            yield resp


class BaseEmbeddingLanguageModel(BaseLanguageModel[Embedding]):
    input: Union[str, List[str]]
    encoding_format: Optional[str] = None
    user: Optional[str] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @task()
    async def _arun(
        self,
        endpoint_id: str,
        request: MaasEmbeddingsRequest,
    ) -> MaasEmbeddingsResponse:
        return await self.client.async_embeddings(endpoint_id=endpoint_id, req=request)

    @task()
    def _run(
        self, endpoint_id: str, request: MaasEmbeddingsRequest
    ) -> MaasEmbeddingsResponse:
        return self.client.embeddings(endpoint_id=endpoint_id, req=request.model_dump())

    async def arun(self, *args: Any, **kwargs: Any) -> MaasEmbeddingsResponse:
        input = [self.input] if isinstance(self.input, str) else self.input

        request = MaasEmbeddingsRequest(
            input=input, encoding_format=self.encoding_format, user=self.user, **kwargs
        )

        return await self._arun(self.endpoint_id, request)

    def run(self, *args: Any, **kwargs: Any) -> MaasEmbeddingsResponse:
        input = [self.input] if isinstance(self.input, str) else self.input

        request = MaasEmbeddingsRequest(
            input=input, encoding_format=self.encoding_format, user=self.user, **kwargs
        )

        return self._run(self.endpoint_id, request)

    def batch(self, *args: Any, **kwargs: Any) -> List[Embedding]:
        assert isinstance(self.input, list)

        request = MaasEmbeddingsRequest(
            input=self.input,
            encoding_format=self.encoding_format,
            user=self.user,
            **kwargs,
        )

        resp = self._run(self.endpoint_id, request)
        return resp.data

    async def abatch(self, *args: Any, **kwargs: Any) -> List[Embedding]:
        assert isinstance(self.input, list)

        request = MaasEmbeddingsRequest(
            input=self.input,
            encoding_format=self.encoding_format,
            user=self.user,
            **kwargs,
        )

        resp = await self._arun(self.endpoint_id, request)
        return resp.data


class BaseTokenizeLanguageModel(BaseLanguageModel):
    endpoint_id: str
    text: str

    @task()
    async def _arun(
        self,
        endpoint_id: str,
        request: MaasTokenizeRequest,
    ) -> MaasTokenizeResponse:
        return await self.client.async_tokenize(endpoint_id, req=request)

    @task()
    def _run(
        self, endpoint_id: str, request: MaasTokenizeRequest
    ) -> MaasTokenizeResponse:
        return self.client.tokenize(endpoint_id, req=request.model_dump())

    async def arun(self, *args: Any, **kwargs: Any) -> MaasTokenizeResponse:
        request = MaasTokenizeRequest(text=self.text)

        return await self._arun(self.endpoint_id, request)

    def run(self, *args: Any, **kwargs: Any) -> MaasTokenizeResponse:
        request = MaasTokenizeRequest(text=self.text)

        return self._run(self.endpoint_id, request)


class BaseClassificationLanguageModel(BaseLanguageModel):
    query: str
    labels: List[str]

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @task()
    async def _arun(
        self,
        endpoint_id: str,
        request: MaasClassificationRequest,
    ) -> MaasClassificationResponse:
        return await self.client.async_classification(
            endpoint_id=endpoint_id, req=request
        )

    @task()
    def _run(
        self,
        endpoint_id: str,
        request: MaasClassificationRequest,
    ) -> MaasClassificationResponse:
        return self.client.classification(
            endpoint_id=endpoint_id, req=request.model_dump()
        )

    async def arun(self, *args: Any, **kwargs: Any) -> Any:
        request = MaasClassificationRequest(
            query=self.query, labels=self.labels, **kwargs
        )

        return await self._arun(self.endpoint_id, request)

    def run(self, *args: Any, **kwargs: Any) -> Any:
        request = MaasClassificationRequest(
            query=self.query, labels=self.labels, **kwargs
        )

        return self._run(self.endpoint_id, request)



from typing import AsyncIterable

from ark.core.idl.maas_protocol import MaasChatRequest, MaasChatResponse
from ark.core.launcher.local.serve import launch_serve
from ark.core.task import task


@task(distributed=False)
async def main(request: MaasChatRequest) -> AsyncIterable[MaasChatResponse]:
    # Your code here
    response = MaasChatResponse()
    yield response


if __name__ == "__main__":
    launch_serve(package_path="main")



import os
from enum import Enum
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Coroutine,
    Dict,
    Optional,
    Tuple,
    Type,
    Union,
)

from pydantic import ValidationError
from volcenginesdkarkruntime._exceptions import ArkAPIError

from ark.core.client import Client
from ark.core.idl.common_protocol import (
    Context,
    Request,
    Response,
)
from ark.core.launcher.runner import get_endpoint_config, get_runner
from ark.core.launcher.vefaas.common import parse_request, parse_response
from ark.core.launcher.vefaas.initializer import initialize
from ark.core.utils.errorsv3 import (
    APIException,
    InternalServiceError,
    MissingParameter,
)
from ark.core.utils.types import parse_pydantic_error
from ark.telemetry.trace import TraceConfig

StreamingResponse = AsyncIterator[str]
JsonResponse = Dict[str, Any]


class Environment(str, Enum):
    VEFAAS = "VeFaaS"
    LOCAL = "Local"


def _get_parameters(
    environment: Environment, *args: Any, **kwargs: Any
) -> Tuple[Union[Dict[str, Any], Request], Any]:
    if environment == Environment.LOCAL:
        if len(args) == 0:
            raise MissingParameter("request")
        return args[0], Context()
    else:
        if len(args) == 2:
            return args[0], args[1]
        if len(kwargs) == 2:
            event, context = kwargs.get("event", ""), kwargs.get("context", "")
            assert event and context, "No valid faas input"

            return event, context

        raise TypeError("No valid faas input")


def parse_function_request(
    environment: Environment,
    parameters: Union[Dict[str, Any], Request],
    endpoint_path: str,
    func: Callable,
) -> Request:
    # load request class type by signature
    endpoint_config = get_endpoint_config(
        endpoint_path=endpoint_path, runnable_func=func
    )
    request_cls = endpoint_config.get(endpoint_path, Request)

    if environment == Environment.LOCAL:
        # load args as it looks
        request: Request = (
            request_cls.model_validate(parameters)
            if isinstance(parameters, dict)
            else parameters
        )
    else:
        # transform the input args for vefaas deployment
        request = parse_request(
            event=parameters,
            request_cls=request_cls,
        )
    return request


def parse_function_response(
    environment: Environment,
    *,
    response: Optional[Response] = None,
    exception: Optional[APIException] = None,
) -> Union[Response, JsonResponse]:
    if environment == Environment.LOCAL:
        if exception:
            # error should be directly raised
            raise exception
        elif response:
            # response should be directly returned
            return response
    else:
        if exception:
            # error should return with http code
            return parse_response(
                status_code=exception.http_code,
                content=Response(error=exception.to_error()).model_dump_json(
                    exclude_none=True
                ),
            )
        elif response:
            # response should return with serialization
            return parse_response(
                status_code=200,
                content=response.model_dump_json(exclude_none=True),
            )
    return Response()


def bot_wrapper(
    endpoint_path: str = "/api/v3/bots/chat/completions",
    clients: Optional[Dict[str, Tuple[Type[Client], Any]]] = None,
    trace_on: bool = True,
    trace_config: Optional[TraceConfig] = None,
) -> Any:
    """
    This bot_wrapper is for launching vefaas function,
    to initialize the environment and translate request/response.
    Args:
        endpoint_path: The api path of the procode service for client to request
            e.g. https://{host}/api/v3/bots/chat/completions
        clients: The clients for the procode service,
                 each key stands for client name,
                 each value stands for the client initialize params.
            e.g. {
                "ark": (
                    AsyncArk,
                    {
                        "base_url": "https://ark.cn-beijing.volces.com/api/v3",
                        "max_retries": 3,
                    },
            )}
        trace_on: The option stands for using tracing, ref: https://opentelemetry.io/docs/
        if turn on, with @task decorator represents tracing
        the duration & input/output of the decorated function.
        trace_config: Detailed trace config for tracing
    """

    def wrapper(
        func: Callable,
    ) -> Callable[
        [Tuple[Any, ...], Dict[str, Any]],
        Union[
            Coroutine[Any, Any, Union[JsonResponse, StreamingResponse, Response]],
            JsonResponse,
        ],
    ]:
        def run_bot(
            *args: Any, **kwargs: Any
        ) -> Union[
            Coroutine[Any, Any, Union[JsonResponse, StreamingResponse, Response]],
            JsonResponse,
        ]:
            environment: Environment = Environment.VEFAAS
            if os.getenv("IS_LOCAL") is not None:
                environment = Environment.LOCAL

            # decode environment-related parameters
            parameters, context = _get_parameters(environment, *args, **kwargs)

            # initialize environment
            initialize(context, clients, trace_on, trace_config)

            # get runner for handler(handle sse & error code)
            runner = get_runner(func)
            try:
                # encode request
                request = parse_function_request(
                    environment, parameters, endpoint_path, func
                )
            except Exception as e:
                if isinstance(e, APIException):
                    runtime_exception: APIException = e
                elif isinstance(e, ValidationError):
                    runtime_exception = parse_pydantic_error(e)
                else:
                    runtime_exception = InternalServiceError(str(e))

                return parse_response(
                    status_code=runtime_exception.http_code,
                    content=Response(
                        error=runtime_exception.to_error()
                    ).model_dump_json(exclude_none=True),
                )

            async def entry() -> Union[JsonResponse, StreamingResponse, Response]:
                try:
                    # run function
                    if request.stream:
                        return runner.astream(request)
                    else:
                        response = await runner.arun(request)

                        return parse_function_response(environment, response=response)
                except Exception as e:
                    if isinstance(e, APIException):
                        runtime_exception: APIException = e
                    elif isinstance(e, ArkAPIError):
                        runtime_exception = APIException(
                            http_code=e.status_code
                            if hasattr(e, "status_code")
                            else 500,
                            code=e.code,
                            message=e.message,
                            parameter=e.param,
                            error_type=e.type,
                        )
                    elif isinstance(e, ValidationError):
                        runtime_exception = parse_pydantic_error(e)
                    else:
                        runtime_exception = InternalServiceError(str(e))

                    return parse_function_response(
                        environment, exception=runtime_exception
                    )

            return entry()

        return run_bot

    return wrapper



from typing import Any, Dict, Tuple, Type, Union

from pydantic import Field

from ark.core.action import Action
from ark.core.idl.ark_protocol import (
    ArkCalculatorRequest,
    ArkCalculatorResponse,
    FunctionDefinition,
)
from ark.core.idl.common_protocol import ActionDetails
from ark.core.task.task import task


def _get_calculator_schema() -> FunctionDefinition:
    return FunctionDefinition(
        name="calculator",
        description="Evaluate a given mathematical expression",
        parameters={
            "type": "object",
            "properties": {
                "input": {
                    "type": "string",
                    "description": "The mathematical expression "
                    "in Wolfram Language InputForm",
                },
            },
            "required": ["input"],
        },
    )


class Calculator(Action[ArkCalculatorRequest, ArkCalculatorResponse]):
    name: str = "Calculator"
    function_definition: FunctionDefinition = Field(
        default_factory=_get_calculator_schema
    )
    response_cls: Type[ArkCalculatorResponse] = ArkCalculatorResponse

    @task()
    async def arun(
        self, request: ArkCalculatorRequest, **kwargs: Any
    ) -> Union[ArkCalculatorResponse, Tuple[ArkCalculatorResponse, ActionDetails]]:
        return await super().arun(request, **kwargs)

    @task()
    async def acall(
        self, request: Dict[str, Any], **kwargs: Any
    ) -> Union[ArkCalculatorResponse, Tuple[ArkCalculatorResponse, ActionDetails]]:
        """
        for function call format
        """
        req = ArkCalculatorRequest.model_validate(request)
        return await super().arun(req, **kwargs)



from abc import abstractmethod
from typing import Any, Optional, TypeVar, Union

from langchain_core.output_parsers import BaseOutputParser
from langchain_core.prompts import BasePromptTemplate
from pydantic.v1 import BaseModel, Field
from volcenginesdkarkruntime import AsyncArk

from ark.core.client import ArkClient, get_client_pool
from ark.core.runnable import Runnable


def _default_ark_client() -> ArkClient:
    client_pool = get_client_pool()
    client: ArkClient = client_pool.get_client("chat")  # type: ignore
    if not client:
        client = ArkClient()
    return client


T = TypeVar("T")


class BaseLanguageModel(BaseModel, Runnable[T]):
    endpoint_id: str
    client: Union[ArkClient, AsyncArk] = Field(default_factory=_default_ark_client)
    template: Optional[BasePromptTemplate] = None
    output_parser: Optional[BaseOutputParser] = None

    class Config:
        arbitrary_types_allowed = True

    @abstractmethod
    def _run(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError

    @abstractmethod
    async def _arun(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError



from importlib import metadata

from ark import component, core
from ark.core import idl, launcher

__all__ = ["core", "component", "idl", "launcher"]

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""



from typing import Any, Dict, List, Optional, Union

from pydantic import Field, PrivateAttr, field_validator

from ark.component.plugins.plugin import Plugin, convert_messages_to_keywords
from ark.core._api.deprecation import deprecated
from ark.core.client import ArkClient
from ark.core.idl.common_protocol import Parameters, Tool
from ark.core.idl.maas_protocol import (
    MaasChatChoice,
    MaasChatMessage,
)
from ark.core.utils.errors import InvalidParameter, MissingParameter


def _default_result_mappings() -> Dict[str, bool]:
    return {"需要": True, "不需要": False}


@deprecated(
    since="0.1.11",
    removal="0.2.0",
    alternative_import="ark.component.v3.search.SearchIntention",
)
class SearchIntention(Plugin):
    _tools: Optional[List[Tool]] = None
    _result_mappings: Optional[Dict[str, bool]] = PrivateAttr(
        default_factory=_default_result_mappings
    )

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def __init__(
        self,
        endpoint_id: str,
        messages: List[MaasChatMessage] = Field(default_factory=list),
        options: Optional[Dict[str, Any]] = None,
        parameters: Optional[Union[Parameters, Dict[str, Any]]] = None,
        client: Optional[ArkClient] = None,
        **kwargs: Any,
    ):
        super().__init__(
            name="SearchIntention",
            endpoint_id=endpoint_id,
            messages=messages,
            options=options,
            parameters=parameters,
            **kwargs,
        )
        if client:
            self.client = client

        if options:
            if "result_mapping" in options and isinstance(
                options["result_mapping"], dict
            ):
                self._result_mappings = (
                    options["result_mapping"] or _default_result_mappings()
                )

            keywords = options.get("keywords", [])

            if len(keywords) > 0 and all(
                isinstance(item, MaasChatMessage) for item in keywords
            ):
                options["keywords"] = convert_messages_to_keywords(keywords)

        self._tools = [Tool(type=self.name, options=options)]

    async def aparse_output(self, output: Union[str, List[MaasChatChoice]]) -> bool:
        _result_mappings = self._result_mappings or _default_result_mappings()
        text = output if isinstance(output, str) else ""
        if isinstance(output, list):
            if len(output) == 0:
                return False
            text = (
                output[0].message.content
                if isinstance(output[0].message.content, str)
                else ""
            )

        # default false
        return _result_mappings.get(text, False)

    @field_validator("options")
    @classmethod
    def validate_options(cls, v: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
        if not v or "keywords" not in v:
            return v

        keywords = v.get("keywords", [])

        if keywords and not isinstance(keywords, list):
            raise InvalidParameter("keywords should be list")

        # no keywords, simply return, result would be parsed by
        if len(keywords) == 0:
            return v

        if any(
            not isinstance(item, str) and not isinstance(item, MaasChatMessage)
            for item in keywords
        ):
            raise InvalidParameter(
                "keywords should be list of either str or MaasChatMessage"
            )

        if "result_mapping" not in v or not isinstance(v["result_mapping"], dict):
            raise MissingParameter(
                "options should contain result_mapping when keywords exist"
            )

        return v



import logging
import time
from dataclasses import dataclass, field
from typing import Callable

from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from ark.core.utils.context import (
    get_client_reqid,
    get_reqid,
    get_start_time,
    set_client_reqid,
    set_headers,
    set_reqid,
    set_start_time,
)
from ark.core.utils.logid import gen_log_id


@dataclass
class LogIdMiddleware:
    app: "ASGIApp"
    client_header_name: str = "x-client-request-id"
    header_name: str = "x-request-id"

    generator: Callable[[], str] = field(default=lambda: gen_log_id())

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Load log ID from headers if present. Generate one otherwise.
        And put it into context.
        """
        set_start_time(time.perf_counter())
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        headers = MutableHeaders(scope=scope)

        headers[self.header_name] = self.generator()
        headers[self.client_header_name] = headers.get(
            self.client_header_name.lower(), headers[self.header_name]
        )

        set_reqid(headers[self.header_name])
        set_client_reqid(headers[self.client_header_name])
        set_headers(headers)

        async def handle_outgoing_request(message: "Message") -> None:
            if message["type"] == "http.response.start" and get_reqid():
                headers = MutableHeaders(scope=message)
                headers.append(self.header_name, get_reqid())
                headers.append(self.client_header_name, get_client_reqid())

            await send(message)
            logging.debug(
                f"[{get_reqid()}] out app cost={time.perf_counter() - get_start_time()}"
            )

        await self.app(scope, receive, handle_outgoing_request)
        return



from typing import Any, Dict, Tuple, Type, Union

from pydantic import Field

from ark.core.action import Action
from ark.core.idl.ark_protocol import (
    ArkLinkReaderRequest,
    ArkLinkReaderResponse,
    FunctionDefinition,
)
from ark.core.idl.common_protocol import ActionDetails
from ark.core.task.task import task
from ark.core.utils.errorsv3 import InvalidParameter


def get_link_reader_schema() -> FunctionDefinition:
    return FunctionDefinition(
        name="LinkReader",
        description="当你需要获取网页、pdf、抖音视频内容时，使用此工具。可以获取url链接下的标题和内容。"
        '\n\nexamples: [{"url_list": ["abc.com", "xyz.com"]}]',
        parameters={
            "type": "object",
            "properties": {
                "url_list": {"type": "array", "items": {"type": "string"}},
                "description": "需要解析网页链接,最多3个,以列表返回",
            },
            "required": ["url_list"],
        },
    )


class LinkReader(Action[ArkLinkReaderRequest, ArkLinkReaderResponse]):
    name: str = "LinkReader"
    function_definition: FunctionDefinition = Field(
        default_factory=get_link_reader_schema
    )
    response_cls: Type[ArkLinkReaderResponse] = ArkLinkReaderResponse

    @task()
    async def arun(
        self, request: ArkLinkReaderRequest, **kwargs: Any
    ) -> Union[ArkLinkReaderResponse, Tuple[ArkLinkReaderResponse, ActionDetails]]:
        return await super().arun(request, **kwargs)

    @task()
    def _fallback_dict_param(self, dict_param: Dict[str, Any]) -> ArkLinkReaderRequest:
        if "url_list" in dict_param:
            return ArkLinkReaderRequest(url_list=dict_param["url_list"])

        if len(dict_param) != 1:
            raise InvalidParameter(parameter="url_list")

        item = list(dict_param.values())[0]
        if isinstance(item, list):
            return ArkLinkReaderRequest(url_list=item)
        else:
            return ArkLinkReaderRequest(url_list=[str(item)])

    @task()
    async def acall(
        self, request: Dict[str, Any], **kwargs: Any
    ) -> Union[ArkLinkReaderResponse, Tuple[ArkLinkReaderResponse, ActionDetails]]:
        """
        for function call format
        """
        req = self._fallback_dict_param(request)
        return await super().arun(req, **kwargs)



import copy
import json
import logging
from typing import Any, Dict, List, Optional, Tuple, Union

from httpx import Timeout
from langchain.prompts.chat import BaseChatPromptTemplate
from langchain.schema.output_parser import BaseTransformOutputParser
from pydantic import BaseModel
from pydantic.v1 import Field
from typing_extensions import Literal
from volcenginesdkarkruntime import Ark, AsyncArk
from volcenginesdkarkruntime._streaming import AsyncStream
from volcenginesdkarkruntime.types.chat import (
    ChatCompletion,
    ChatCompletionChunk,
)
from volcenginesdkarkruntime.types.create_embedding_response import (
    CreateEmbeddingResponse,
)
from volcenginesdkarkruntime.types.create_tokenization_response import (
    CreateTokenizationResponse,
    Tokenization,
)

from ark.core.client import get_client_pool
from ark.core.idl.ark_protocol import (
    ArkChatCompletionChunk,
    ArkChatParameters,
    ArkChatRequest,
    ArkChatResponse,
    ArkMessage,
    CallableFunction,
    EndpointId,
    FunctionCallMode,
)
from ark.core.idl.common_protocol import ActionDetails, BotUsage
from ark.core.llm import BaseLanguageModel
from ark.core.task import task
from ark.core.utils.context import get_extra_headers
from ark.core.utils.prompt import format_ark_prompts
from ark.core.utils.types import convert_response_message, transform_response


def _default_ark_client() -> AsyncArk:
    client_pool = get_client_pool()
    client: AsyncArk = client_pool.get_client("ark")  # type: ignore
    if not client:
        client = AsyncArk(timeout=Timeout(connect=1.0, timeout=60.0))
    return client


class BaseChatLanguageModel(
    BaseLanguageModel[Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]]
):
    client: AsyncArk = Field(default_factory=_default_ark_client)
    messages: List[ArkMessage]
    parameters: Optional[ArkChatParameters] = None
    template: Optional[BaseChatPromptTemplate] = None
    output_parser: Optional[BaseTransformOutputParser] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def generate_prompts(
        self,
        messages: List[ArkMessage],
        *,
        additional_system_prompts: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[ArkMessage]:
        # additional system prompts would be put at first
        if additional_system_prompts:
            prompts = [
                ArkMessage(role="system", content=system_prompt)
                for system_prompt in additional_system_prompts
            ]
            prompts.extend(messages)
            messages = prompts

        if not self.template:
            return messages

        return format_ark_prompts(self.template, messages, **kwargs)

    def get_request_model(self, **kwargs: Any) -> EndpointId:
        return self.endpoint_id

    @task()
    def parse_output(self, text: str) -> Any:
        if not self.output_parser:
            return text

        return self.output_parser.parse(text)

    @task()
    async def aparse_output(self, text: str) -> Any:
        if not self.output_parser:
            return text

        return await self.output_parser.aparse(text)

    @task()
    async def _arun(
        self,
        request: ArkChatRequest,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        assert isinstance(self.client, AsyncArk), TypeError("Invalid Client for v3 sdk")

        params = request.get_chat_request(extra_body)

        extra_headers = get_extra_headers(extra_headers)

        return await self.client.chat.completions.create(
            **params,
            extra_headers=extra_headers,
            extra_query=extra_query,
        )

    @task()
    def _run(
        self,
        request: ArkChatRequest,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        sync_client = Ark()

        extra_headers = get_extra_headers(extra_headers)

        return sync_client.chat.completions.create(
            **request.get_chat_request(extra_body),
            extra_headers=extra_headers,
            extra_query=extra_query,
        )

    def run(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=False,
            messages=self.generate_prompts(self.messages, **kwargs),
            model=self.get_request_model(**kwargs),
            **parameters,
        )

        completion = self._run(request, extra_headers, extra_query, extra_body)
        return ArkChatResponse(**completion.__dict__)

    def stream(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=True,
            messages=self.generate_prompts(self.messages, **kwargs),
            model=self.get_request_model(**kwargs),
            **parameters,
        )

        completion = self._run(request, extra_headers, extra_query, extra_body)
        for resp in completion:
            yield ArkChatCompletionChunk(**resp.__dict__)

    async def handle_function_call(
        self,
        request: ArkChatRequest,
        response: Union[
            ChatCompletionChunk, ChatCompletion, ArkChatCompletionChunk, ArkChatResponse
        ],
        available_functions: Optional[Dict[str, CallableFunction]] = None,
        function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
    ) -> bool:
        return await handle_function_call(
            request, response, available_functions, function_call_mode
        )

    async def arun(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        *,
        available_functions: Optional[Dict[str, CallableFunction]] = None,
        function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
        additional_system_prompts: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> ArkChatResponse:
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=False,
            messages=self.generate_prompts(
                self.messages,
                additional_system_prompts=additional_system_prompts,
                **kwargs,
            ),
            model=self.get_request_model(**kwargs),
            **parameters,
        )
        responses = []
        while True:
            completion: ChatCompletion = await self._arun(
                request, extra_headers, extra_query, extra_body
            )
            responses.append(completion)

            if completion.choices and completion.choices[0].finish_reason:
                if not await self.handle_function_call(
                    request, completion, available_functions, function_call_mode
                ):
                    break

        return ArkChatResponse.merge(responses)

    async def astream(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
        *,
        available_functions: Optional[Dict[str, CallableFunction]] = None,
        function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
        additional_system_prompts: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> AsyncStream[ArkChatCompletionChunk]:
        parameters = (
            self.parameters.model_dump(exclude_none=True, exclude_unset=True)
            if self.parameters
            else {}
        )
        request = ArkChatRequest(
            stream=True,
            messages=self.generate_prompts(
                self.messages,
                additional_system_prompts=additional_system_prompts,
                **kwargs,
            ),
            model=self.get_request_model(**kwargs),
            **parameters,
        )

        usage_chunks, cumulated = [], []
        while True:
            completion = await self._arun(
                request, extra_headers, extra_query, extra_body
            )
            # default: one iter
            is_more_request = False
            async for resp in completion:  # type: ChatCompletionChunk
                if resp.usage:
                    usage_chunks.append(resp)
                else:
                    # cumulated chunks is used for caculator/fc inner cot output
                    cumulated.append(resp)
                    yield ArkChatCompletionChunk(**resp.__dict__)

                if resp.choices and resp.choices[0].finish_reason:
                    ark_resp = ArkChatCompletionChunk(**resp.__dict__).merge(cumulated)
                    # clear after used
                    cumulated = []
                    is_more_request = await self.handle_function_call(
                        request, ark_resp, available_functions, function_call_mode
                    )
                    if (
                        request.stream_options
                        and request.stream_options.get("include_usage") is True
                    ):
                        usage_chunks.append(ark_resp)

            if not is_more_request:
                break

        if len(usage_chunks) > 0:
            yield ArkChatCompletionChunk.merge(usage_chunks)


@task()
async def handle_function_call(
    request: ArkChatRequest,
    response: Union[
        ChatCompletionChunk, ChatCompletion, ArkChatCompletionChunk, ArkChatResponse
    ],
    available_functions: Optional[Dict[str, CallableFunction]] = None,
    function_call_mode: Optional[FunctionCallMode] = FunctionCallMode.SEQUENTIAL,
    **kwargs: Any,
) -> bool:
    if response.choices[0].finish_reason != "tool_calls":
        return False

    response_message = (
        response.choices[0].delta
        if isinstance(response, ChatCompletionChunk)
        or isinstance(response, ArkChatCompletionChunk)
        else response.choices[0].message
    )
    tool_calls = response_message.tool_calls

    if not tool_calls or not available_functions:
        return False

    function_call_mode = function_call_mode or FunctionCallMode.SEQUENTIAL
    if function_call_mode != FunctionCallMode.SEQUENTIAL:
        raise NotImplementedError("Only sequential function call mode is supported")

    request.messages.append(convert_response_message(response_message))

    function_calls = copy.deepcopy(tool_calls)
    for tool_call in function_calls:
        function_name = tool_call.function.name

        function_to_call = available_functions.get(function_name)
        original_function_response: Union[
            Union[str, BaseModel], Tuple[Union[str, BaseModel], ActionDetails]
        ] = ""
        if function_to_call:
            function_args = json.loads(tool_call.function.arguments)
            original_function_response = await function_to_call(function_args, **kwargs)
        else:
            logging.error(f"Function {function_name} not found")

        if (
            isinstance(original_function_response, tuple)
            and len(original_function_response) == 2
            and isinstance(original_function_response[1], ActionDetails)
        ):
            function_response, action_detail = original_function_response
            if isinstance(response, ArkChatCompletionChunk):
                response.bot_usage = BotUsage(action_details=[action_detail])
        else:
            function_response = original_function_response

        request.messages.append(
            ArkMessage(
                role="tool",
                content=transform_response(function_response),
                tool_call_id=tool_call.id,
            )
        )
    return True


class BaseEmbeddingLanguageModel(BaseLanguageModel[List[float]]):
    client: AsyncArk = Field(default_factory=_default_ark_client)
    input: Union[str, List[str]]
    encoding_format: Literal["float", "base64"] = "float"
    user: Optional[str] = None

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    @task()
    async def _arun(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateEmbeddingResponse:
        extra_headers = get_extra_headers(extra_headers)

        return await self.client.embeddings.create(
            model=self.endpoint_id,
            input=self.input,
            encoding_format=self.encoding_format,
            user=self.user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
        )

    @task()
    def _run(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateEmbeddingResponse:
        sync_client = Ark()

        extra_headers = get_extra_headers(extra_headers)

        return sync_client.embeddings.create(
            model=self.endpoint_id,
            input=self.input,
            encoding_format=self.encoding_format,
            user=self.user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
        )

    async def arun(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateEmbeddingResponse:
        return await self._arun(extra_headers, extra_query, extra_body)

    def run(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateEmbeddingResponse:
        return self._run(extra_headers, extra_query, extra_body)

    def batch(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> List[List[float]]:
        resp: CreateEmbeddingResponse = self._run(
            extra_headers, extra_query, extra_body
        )
        embeddings = []
        for data in resp.data:
            embeddings.append(data.embedding)
        return embeddings

    async def abatch(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> List[List[float]]:
        resp: CreateEmbeddingResponse = await self._arun(
            extra_headers, extra_query, extra_body
        )
        embeddings = []
        for data in resp.data:
            embeddings.append(data.embedding)
        return embeddings


class BaseTokenizeLanguageModel(BaseLanguageModel[Tokenization]):
    client: AsyncArk = Field(default_factory=_default_ark_client)
    text: Union[str, List[str]]
    user: Optional[str] = None

    @task()
    async def _arun(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateTokenizationResponse:
        extra_headers = get_extra_headers(extra_headers)

        return await self.client.tokenization.create(
            model=self.endpoint_id,
            text=self.text,
            user=self.user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
        )

    @task()
    def _run(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateTokenizationResponse:
        sync_client = Ark()

        extra_headers = get_extra_headers(extra_headers)

        return sync_client.tokenization.create(
            model=self.endpoint_id,
            text=self.text,
            user=self.user,
            extra_headers=extra_headers,
            extra_query=extra_query,
            extra_body=extra_body,
        )

    async def arun(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateTokenizationResponse:
        return await self._arun(extra_headers, extra_query, extra_body)

    def run(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> CreateTokenizationResponse:
        return self._run(extra_headers, extra_query, extra_body)

    def batch(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> List[Tokenization]:
        resp: CreateTokenizationResponse = self._run(
            extra_headers, extra_query, extra_body
        )
        return resp.data

    async def abatch(
        self,
        extra_headers: Optional[Dict[str, str]] = {},
        extra_query: Optional[Dict[str, Any]] = None,
        extra_body: Optional[Dict[str, Any]] = None,
    ) -> List[Tokenization]:
        resp: CreateTokenizationResponse = await self._arun(
            extra_headers, extra_query, extra_body
        )
        return resp.data
